import base64
import gzip
import io
import json
import logging
import os
import time
from dataclasses import dataclass, field
from json import JSONDecodeError
from typing import TYPE_CHECKING, Optional, Any, Callable, Dict, Iterable, Union, Generator, Iterator

import dill
import pyarrow
from databricks.sdk.errors import InternalError
from databricks.sdk.service.compute import (
    Language, CommandExecutionAPI, CommandStatusResponse, CommandStatus, ResultType
)

from .exceptions import ClientTerminatedSession
from ...pyutils.exceptions import raise_parsed_traceback
from ...pyutils.waiting_config import WaitingConfig, WaitingConfigArg

DONE_STATES = {
    CommandStatus.FINISHED, CommandStatus.CANCELLED, CommandStatus.ERROR
}

PENDING_STATES = {
    CommandStatus.RUNNING, CommandStatus.QUEUED, CommandStatus.RUNNING
}

FAILED_STATES = {
    CommandStatus.ERROR, CommandStatus.CANCELLED
}

if TYPE_CHECKING:
    from .execution_context import ExecutionContext


__all__ = [
    "CommandExecution"
]


LOGGER = logging.getLogger(__name__)


@dataclass
class CommandExecution:
    context: "ExecutionContext"
    command_id: Optional[str] = None

    language: Optional[Language] = field(default=None, repr=False, compare=False, hash=False)
    command: Optional[str] = field(default=None, repr=False, compare=False, hash=False)
    environ: Optional[Dict[str, str]] = field(default=None, repr=False, compare=False, hash=False)

    _details: Optional[CommandStatusResponse] = field(default=None, repr=False, compare=False, hash=False)

    def __post_init__(self):
        if self.environ:
            if isinstance(self.environ, (list, tuple, set)):
                self.environ = {
                    k: os.getenv(k)
                    for k in self.environ
                }

    def __call__(self, *args, **kwargs):
        assert self.command, "Cannot call %s, missing command" % self

        args_blob = dill.dumps([self.encode_object(_) for _ in args])
        kwargs_blob = dill.dumps({k: self.encode_object(v) for k, v in kwargs.items()})

        if self.environ:
            env_blob = {
                k: os.getenv(k) or v
                for k, v in self.environ.items()
                if os.getenv(k) or v
            }
        else:
            env_blob = {}

        args_b64 = base64.b64encode(args_blob).decode("ascii")
        kwargs_b64 = base64.b64encode(kwargs_blob).decode("ascii")

        command = (
            self.command
            .replace("__ARGS__", repr(args_b64))
            .replace("__KWARGS__", repr(kwargs_b64))
            .replace("__ENVIRON__", repr(env_blob))
        )

        run = (
            self.create(
                context=self.context,
                command=command,
                language=self.language
            )
            .start()
        )

        return run.wait(raise_error=True).result(raise_error=True)

    def __bool__(self):
        return self.done

    def __repr__(self):
        return "%s(url=%s)" % (
            self.__class__.__name__,
            self.url()
        )

    def __str__(self):
        return self.url()

    def url(self) -> str:
        return "%s/command/%s" % (
            self.context.url(),
            self.command_id or "unknown"
        )

    def create(
        self,
        context: Optional["ExecutionContext"] = None,
        func: Optional[Callable] = None,
        command: Optional[str] = None,
        language: Optional[Language] = None,
        command_id: Optional[str] = None,
        environ: Optional[Union[Iterable[str], Dict[str, str]]] = None,
    ):
        context = self.context if context is None else context
        command = self.command if command is None else command
        environ = self.environ if environ is None else environ

        if environ is not None:
            if not isinstance(environ, dict):
                environ = {
                    str(k): os.getenv(str(k))
                    for k in environ
                }
            else:
                environ = {
                    str(k): str(v)
                    for k, v in environ.items()
                }

        if not command:
            if callable(func):
                command = self.make_python_function_command(
                    func=func,
                )

        if language is None:
            language = context.language or Language.PYTHON

        assert context is not None, "Missing context to execute command"

        return CommandExecution(
            context=context,
            language=language,
            command=command,
            command_id=command_id,
            environ=environ
        )

    def start(self, reset: bool = False):
        if self.command_id:
            if not reset:
                return self

            self._details = None
            self.command_id = None

        client = self.context.workspace_client().command_execution

        assert self.command, "Missing command arg in %s" % self

        try:
            details = client.execute(
                cluster_id=self.cluster_id,
                context_id=self.context_id,
                language=self.language,
                command=self.command,
            ).response
        except Exception as e:
            if "ontext" in str(e):  # context related
                self.context = self.context.connect(reset=True)

                details = client.execute(
                    cluster_id=self.cluster_id,
                    context_id=self.context_id,
                    language=self.language,
                    command=self.command,
                ).response
            else:
                raise e

        self.command_id = details.id
        self._details = None

        LOGGER.debug("Started %s", self)

        return self

    @property
    def workspace(self):
        return self.context.workspace

    @property
    def cluster_id(self):
        return self.context.cluster.cluster_id

    @property
    def context_id(self):
        if not self.context.context_id:
            self.context = self.context.connect()
        return self.context.context_id

    @property
    def state(self):
        return self.details.status

    @property
    def running(self):
        return self.state in PENDING_STATES

    @property
    def done(self):
        return self.state in DONE_STATES

    @property
    def details(self) -> CommandStatusResponse:
        if self._details is None:
            self._details = self.client().command_status(
                cluster_id=self.cluster_id,
                context_id=self.context_id,
                command_id=self.command_id
            )
        elif self._details.status not in DONE_STATES:
            self._details = self.client().command_status(
                cluster_id=self.cluster_id,
                context_id=self.context_id,
                command_id=self.command_id
            )
        return self._details

    @details.setter
    def details(self, value: Optional[CommandStatusResponse]):
        self._details = value

        if value is not None:
            assert isinstance(value, CommandStatusResponse), "%s.details must be CommandStatusResponse, got %s" %(
                self,
                type(value)
            )
            self.command_id = value.id

    @property
    def results_metadata(self):
        return self.details.results

    def client(self) -> CommandExecutionAPI:
        return self.context.workspace_client().command_execution

    def connect(self, reset: bool = False):
        self.context = self.context.connect(language=self.language)

        return self

    def cancel(self, raise_error: bool = False):
        if self.command_id:
            try:
                self.client().cancel_and_wait(
                    cluster_id=self.cluster_id,
                    command_id=self.command_id,
                    context_id=self.context_id
                )
            except Exception as e:
                if raise_error:
                    raise e
                LOGGER.exception(e)

    def raise_for_status(self):
        if self.state in FAILED_STATES:
            raise_error_from_response(
                response=self.details,
                language=self.language
            )

        return self

    def wait(
        self,
        wait: Optional[WaitingConfigArg] = True,
        raise_error: bool = True
    ):
        if not self.command_id:
            return self.start().wait(
                wait=wait,
                raise_error=raise_error
            )

        wait = WaitingConfig.check_arg(wait)
        iteration, start = 0, time.time()

        if wait.timeout:
            while self.running:
                wait.sleep(iteration=iteration, start=start)
                iteration += 1

        if raise_error:
            try:
                self.raise_for_status()
            except ModuleNotFoundError as e:
                module_name = e.name

                if module_name and not module_name.startswith("ygg"):
                    self.context.cluster.install_temporary_libraries(
                        libraries=[module_name]
                    )

                    return (
                        self
                        .start(reset=True)
                        .wait(wait=wait, raise_error=raise_error)
                    )
                else:
                    raise e
            except ClientTerminatedSession as e:
                LOGGER.error(
                    "%s aborted: %s",
                    self,
                    e
                )

                self.context = self.context.connect(reset=True)

                return (
                    self
                    .start(reset=True)
                    .wait(wait=wait, raise_error=raise_error)
                )

        return self

    def encode_object(
        self,
        obj: Any,
        byte_limit: int = 32 * 1024,
        byref: Any = None,
        recurse: Any = None,
        compression: Optional[str] = None
    ) -> str:
        buffer = io.BytesIO()

        from ...polars.lib import polars
        from ...pandas.lib import pandas

        if isinstance(obj, pyarrow.Table):
            import pyarrow.parquet as pq

            func = "pyarrow.parquet.read_table"
            extension = "parquet"
            pq.write_table(obj, buffer)

            buffer.seek(0)
            dbx_path = self.workspace.tmp_path(extension=extension)
            dbx_path.write_bytes(buffer)

            return json.dumps({
                "func": func,
                "file": dbx_path.full_path()
            })

        if isinstance(obj, polars.DataFrame):
            func = "polars.read_parquet"
            extension = "parquet"
            obj.write_parquet(buffer)

            buffer.seek(0)
            dbx_path = self.workspace.tmp_path(extension=extension)
            dbx_path.write_bytes(buffer)

            return json.dumps({
                "func": func,
                "file": dbx_path.full_path()
            })
        elif isinstance(obj, pandas.DataFrame):
            func = "pandas.read_parquet"
            extension = "parquet"
            obj.to_parquet(path=buffer)

            buffer.seek(0)
            dbx_path = self.workspace.tmp_path(extension=extension)
            dbx_path.write_bytes(buffer)

            return json.dumps({
                "func": func,
                "cpr": compression,
                "file": dbx_path.full_path()
            })
        elif isinstance(obj, (Generator, Iterator)):
            return json.dumps({
                "func": "generator",
                "items": [
                    self.encode_object(_, byte_limit=byte_limit, byref=byref, recurse=recurse, compression=compression)
                    for _ in obj
                ]
            })

        dill.dump(obj, buffer, byref=byref, recurse=recurse)

        raw = buffer.getvalue()

        if compression or len(raw) > byte_limit:
            compression = compression or "gzip"
            if compression != "gzip":
                raise ValueError(f"Unsupported compression: {compression}")
            raw = gzip.compress(raw)

        if len(raw) > byte_limit:
            dbx_path = self.workspace.tmp_path(extension="bin")
            dbx_path.write_bytes(raw)
            return json.dumps({
                "func": "dill.load",
                "cpr": compression,
                "file": dbx_path.full_path(),
            })

        return json.dumps({
            "func": "dill.load",
            "cpr": compression,
            "b64": base64.b64encode(raw).decode("ascii"),
        })

    def decode_payload(
        self,
        payload: Union[str, bytes, dict, list],
        temporary: bool = True
    ):
        if isinstance(payload, (str, bytes)):
            try:
                payload = json.loads(payload)
            except JSONDecodeError:
                return payload

        if isinstance(payload, dict):
            func, compression, b64, databricks_path = (
                payload.get("func"), payload.get("cpr"),
                payload.get("b64"), payload.get("file")
            )

            if isinstance(func, str) and func:
                if b64:
                    blob = base64.b64decode(b64.encode("ascii"))
                elif databricks_path:
                    blob = self.workspace.dbfs_path(databricks_path, temporary=temporary).read_bytes()
                else:
                    blob = None

                if func == "dill.load":
                    if compression == "gzip":
                        import gzip
                        blob = gzip.decompress(blob)

                    return dill.loads(blob)
                elif func.startswith("pyarrow."):
                    import pyarrow.parquet as pq

                    buff = io.BytesIO(blob)

                    return pq.read_table(buff)
                elif func.startswith("pandas."):
                    import pandas

                    buff = io.BytesIO(blob)

                    if func == "pandas.read_parquet":
                        return pandas.read_parquet(buff)
                    elif func == "pandas.read_pickle":
                        return pandas.read_pickle(buff, compression=compression)
                    else:
                        raise NotImplementedError
                elif func == "generator":
                    items = payload.get("items")

                    def gen(it: Iterator = items):
                        if it:
                            for item in it:
                                yield self.decode_payload(item)

                    return gen()
                elif func.startswith("polars."):
                    import polars

                    buff = io.BytesIO(blob)
                    return polars.read_parquet(buff)
                else:
                    raise NotImplementedError

        return payload

    def make_python_function_command(
        self,
        func: Callable,
        tag: str = "__CALL_RESULT__",
        byref: Any = None,
        recurse: Any = None,
    ):
        # Serialize the command object (self) as ASCII-safe base64
        command_bytes = dill.dumps(self)
        command_b64 = base64.b64encode(command_bytes).decode("ascii")

        # Func serialized by strict encoder: DILL:<compression>:b64:<...> or DATABRICKS_PATH:<compression>:path:<...>
        serialized_func = self.encode_object(
            func,
            byte_limit=64 * 1024,
            byref=byref, recurse=recurse
        )

        cmd = f"""
import base64, dill, os

if __ENVIRON__:
    for k, v in __ENVIRON__.items():
        if k and v:
            os.environ[k] = v

command = dill.loads(base64.b64decode({command_b64!r}.encode("ascii")))
args = dill.loads(base64.b64decode(__ARGS__.encode("ascii")))
kwargs = dill.loads(base64.b64decode(__KWARGS__.encode("ascii")))
f = command.decode_payload({serialized_func!r}, temporary=False)
a = [command.decode_payload(x) for x in args]
kw = {{k: command.decode_payload(v) for k, v in kwargs.items()}}
r = f(*a, **kw)
print({tag!r} + command.encode_object(r))"""

        return cmd

    def decode_response(
        self,
        response: CommandStatusResponse,
        language: Language,
        raise_error: bool = True,
        tag: str = "__CALL_RESULT__",
        logger: bool = True,
        unpickle: bool = True
    ) -> Any:
        """Mirror the old Cluster.execute_command result handling.

        Args:
            response: Raw command execution response.
            language: Language executed
            raise_error: Raise error if response is failed
            tag: Result tag
            logger: Print logs
            unpickle: Unpickle

        Returns:
            The decoded output string.
        """
        raise_error_from_response(
            response=response,
            language=language,
            raise_error=raise_error
        )

        results = response.results

        # normal output
        if results.result_type == ResultType.TEXT:
            data = results.data or ""
        else:
            raise NotImplementedError(
                "Cannot decode result form %s" % response
            )

        raw_result = data

        if tag in raw_result:
            logs_text, raw_result = raw_result.split(tag, 1)

            try:
                if logger:
                    for line in logs_text.splitlines():
                        stripped_log = line.strip()

                        if stripped_log:
                            print(stripped_log)
            except Exception as e:
                LOGGER.warning(
                    "Cannot print logs from %s: %s",
                    logs_text,
                    e
                )

        if unpickle:
            return self.decode_payload(payload=raw_result)
        return raw_result

    def result(
        self,
        raise_error: bool = True,
        unpickle: bool = True
    ) -> Any:
        try:
            self.wait(raise_error=raise_error)

            obj = self.decode_response(
                response=self.details,
                language=self.language,
                raise_error=raise_error,
                unpickle=unpickle
            )
        except (InternalError, ClientTerminatedSession):
            self.context = self.context.connect(reset=True)

            return (
                self
                .start(reset=True)
                .result(raise_error=raise_error, unpickle=unpickle)
            )
        except ModuleNotFoundError as e:
            module_name = e.name

            if module_name and not module_name.startswith("ygg"):
                self.context.cluster.install_temporary_libraries(libraries=[module_name])

                return (
                    self
                    .start(reset=True)
                    .result(raise_error=raise_error, unpickle=unpickle)
                )
            else:
                raise e

        return obj


def raise_error_from_response(
    response: CommandStatusResponse,
    language: Language,
    raise_error: bool = True
):
    if raise_error:
        results = response.results

        if results.result_type == ResultType.ERROR:
            message = results.cause or "Command execution failed"

            if "client terminated the session" in message:
                raise ClientTerminatedSession(message)

            if language == Language.PYTHON:
                raise_parsed_traceback(message)

            raise RuntimeError(str(response))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Mapping, Tuple, Union, Any, Iterable, Optional
from urllib.parse import (
    parse_qsl,
    quote,
    unquote,
    urlencode,
    urljoin,
    urlsplit,
    urlunsplit,
)

from yggdrasil.io.parameters import anonymize_parameters

_DEFAULT_PORTS = {"http": 80, "https": 443, "ws": 80, "wss": 443}

# RFC 3986-ish safe sets
_SAFE_PATH = "/:@-._~!$&'()*+,;="
# Query must NOT escape separators '&' and '=' when query is already structured
_SAFE_QUERY = "-._~!$'()*+,;=:@/?&="
_SAFE_FRAGMENT = "-._~!$&'()*+,;=:@/?"

# port == 0 means "absent"
_NO_PORT = 0


def _lower_if(s: str) -> str:
    return s.lower() if s else ""


def _strip_trailing_dot(host: str) -> str:
    if not host:
        return host
    return host[:-1] if host.endswith(".") else host


def _normalize_path(path: str) -> str:
    if not path:
        return ""
    # collapse repeated slashes (except keep single "/")
    if path != "/" and "//" in path:
        while "//" in path:
            path = path.replace("//", "/")
    return path


def _remove_default_port(scheme: str, host: str, port: int) -> int:
    """
    Return 0 when port is absent or is the scheme default.
    """
    if not scheme or not host or port <= 0:
        return _NO_PORT
    default = _DEFAULT_PORTS.get(scheme)
    return _NO_PORT if default == port else port


def _encode_userinfo(userinfo: str) -> str:
    return quote(userinfo, safe=":!$&'()*+,;=") if userinfo else ""


def _encode_path(path: str) -> str:
    # keep existing '%' to avoid double-encoding
    return quote(path, safe=_SAFE_PATH + "%")


def _encode_query(query: str) -> str:
    # Query is structured: keep '&' and '='; keep '%' to avoid double-encoding
    return quote(query, safe=_SAFE_QUERY + "%")


def _encode_fragment(fragment: str) -> str:
    return quote(fragment, safe=_SAFE_FRAGMENT + "%")


def _decode_maybe(s: str, decode: bool) -> str:
    return unquote(s) if decode and s else s


def _parse_netloc(netloc: str, *, decode: bool) -> tuple[str, str, int]:
    """
    Parse netloc into (userinfo, host, port). Port uses 0 as "absent".
    Handles:
      - userinfo@host:port
      - IPv6: [::1]:8080
      - host without port
    """
    if not netloc:
        return "", "", _NO_PORT

    userinfo = ""
    hostport = netloc

    if "@" in netloc:
        userinfo, hostport = netloc.rsplit("@", 1)
        userinfo = _decode_maybe(userinfo, decode)

    host = ""
    port = _NO_PORT

    if hostport.startswith("["):
        rb = hostport.find("]")
        if rb == -1:
            # malformed, treat whole as host
            return userinfo, hostport, _NO_PORT
        host = hostport[1:rb]
        rest = hostport[rb + 1 :]
        if rest.startswith(":") and len(rest) > 1:
            p = rest[1:]
            if p.isdigit():
                port = int(p)
        return userinfo, host, port

    if ":" in hostport:
        h, p = hostport.rsplit(":", 1)
        host = h
        if p.isdigit():
            port = int(p)
    else:
        host = hostport

    return userinfo, host, port


@dataclass(frozen=True, slots=True)
class URL:
    scheme: str = ""
    userinfo: str = ""
    host: str = ""
    port: int = _NO_PORT  # 0 means absent
    path: str = ""
    query: str = ""
    fragment: str = ""

    @classmethod
    def parse_any(
        cls,
        obj: Any,
        *,
        decode: bool = False,
        normalize: bool = True,
    ):
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, dict):
            return cls.parse_dict(
                obj,
                decode=decode,
                normalize=normalize
            )

        return cls.parse(
            raw=str(obj),
            decode=decode,
            normalize=normalize,
        )

    @classmethod
    def parse(
        cls,
        raw: str,
        *,
        decode: bool = False,
        normalize: bool = True,
    ) -> "URL":
        sp = urlsplit(raw)

        scheme = sp.scheme or ""
        userinfo, host, port = _parse_netloc(sp.netloc, decode=decode)

        path = _decode_maybe(sp.path, decode)
        query = _decode_maybe(sp.query, decode)
        fragment = _decode_maybe(sp.fragment, decode)

        if not normalize:
            return URL(
                scheme=scheme,
                userinfo=userinfo,
                host=host,
                port=port,
                path=path,
                query=query,
                fragment=fragment,
            )

        scheme_n = _lower_if(scheme)
        host_n = _strip_trailing_dot(_lower_if(host))
        path_n = _normalize_path(path)
        port_n = _remove_default_port(scheme_n, host_n, port)

        query_n = query
        if query_n:
            items = parse_qsl(query_n, keep_blank_values=True)
            items.sort(key=lambda kv: (kv[0], kv[1]))
            query_n = urlencode(items, doseq=True)

        return cls(
            scheme=scheme_n,
            userinfo=userinfo,
            host=host_n,
            port=port_n,
            path=path_n,
            query=query_n,
            fragment=fragment,
        )

    @classmethod
    def parse_dict(
        cls,
        d: Mapping[str, Any],
        *,
        decode: bool = False,
        normalize: bool = True,
    ) -> "URL":
        """
        Build a URL from a mapping.

        Supported inputs:
          - {"url": "..."} (delegates to parse)
          - {"scheme","netloc","path","query","fragment"} (like urlsplit result)
          - {"scheme","userinfo","host","port","path","query","fragment"}
          - {"scheme","netloc", ...} plus overrides (userinfo/host/port)

        Notes:
          - port == 0 / None => absent
          - if both netloc and (userinfo/host/port) are provided, explicit fields win
        """
        if not d:
            return cls()

        # 1) Common shortcut: {"url": "..."}
        raw = d.get("url") or d.get("raw")
        if raw is not None:
            return cls.parse(str(raw), decode=decode, normalize=normalize)

        # 2) Pull split-ish fields
        scheme = str(d.get("scheme") or "")
        path = str(d.get("path") or "")
        query = str(d.get("query") or "")
        fragment = str(d.get("fragment") or "")

        # Optional full authority in one string
        netloc = d.get("netloc")
        if netloc is None:
            # Some libs use "authority"
            netloc = d.get("authority")
        netloc_s = str(netloc) if netloc is not None else ""

        # 3) Parse netloc if present
        userinfo = ""
        host = ""
        port = _NO_PORT
        if netloc_s:
            userinfo, host, port = _parse_netloc(netloc_s, decode=decode)

        # 4) Explicit overrides win (for dicts that already have fields split out)
        if "userinfo" in d and d["userinfo"] is not None:
            userinfo = str(d["userinfo"])
            userinfo = _decode_maybe(userinfo, decode)

        if "host" in d and d["host"] is not None:
            host = str(d["host"])
            host = _decode_maybe(host, decode)

        if "port" in d:
            p = d.get("port")
            if p is None or p == "" or p == 0:
                port = _NO_PORT
            elif isinstance(p, int):
                port = p if p > 0 else _NO_PORT
            else:
                ps = str(p)
                port = int(ps) if ps.isdigit() and int(ps) > 0 else _NO_PORT

        # 5) Decode other parts (optionally)
        path = _decode_maybe(path, decode)
        query = _decode_maybe(query, decode)
        fragment = _decode_maybe(fragment, decode)

        if not normalize:
            return cls(
                scheme=scheme,
                userinfo=userinfo,
                host=host,
                port=port,
                path=path,
                query=query.lstrip("?"),
                fragment=fragment.lstrip("#"),
            )

        # 6) Apply the same normalization policy as parse()
        scheme_n = _lower_if(scheme)
        host_n = _strip_trailing_dot(_lower_if(host))
        path_n = _normalize_path(path)
        port_n = _remove_default_port(scheme_n, host_n, port)

        query_n = query.lstrip("?")
        if query_n:
            items = parse_qsl(query_n, keep_blank_values=True)
            items.sort(key=lambda kv: (kv[0], kv[1]))
            query_n = urlencode(items, doseq=True)

        fragment_n = fragment.lstrip("#")

        return cls(
            scheme=scheme_n,
            userinfo=userinfo,
            host=host_n,
            port=port_n,
            path=path_n,
            query=query_n,
            fragment=fragment_n,
        )

    def __str__(self) -> str:
        return self.to_string()

    def __truediv__(self, other):
        if isinstance(other, str):
            return self.with_path(path=other)
        raise ValueError("Cannot join %s with %s" % (self, type(other)))

    @property
    def query_dict(self) -> Mapping[str, Tuple[str, ...]]:
        """
        Parsed query as an immutable mapping: key -> tuple(values...)

        - Preserves original order of values per key (matching parse_qsl order)
        - Empty query => {}
        - Keeps blank values (k=) and keys without '=' (k) as ("",)
        """
        if not self.query:
            return {}

        out: dict[str, list[str]] = {}
        for k, v in parse_qsl(self.query, keep_blank_values=True):
            out.setdefault(k, []).append(v)

        return {k: tuple(vs) for k, vs in out.items()}

    @property
    def is_absolute(self):
        return bool(self.scheme) and bool(self.host)

    def to_string(self, *, encode: bool = True) -> str:
        scheme = self.scheme
        host = self.host
        userinfo = self.userinfo
        path = self.path
        query = self.query
        fragment = self.fragment

        if encode:
            if userinfo:
                userinfo = _encode_userinfo(userinfo)
            path = _encode_path(path)
            query = _encode_query(query)
            fragment = _encode_fragment(fragment)

        netloc = ""
        if host:
            out_host = host
            # bracket IPv6
            if ":" in out_host and not out_host.startswith("["):
                out_host = f"[{out_host}]"
            netloc = out_host

            if self.port > 0:
                netloc = f"{netloc}:{self.port}"

            if userinfo:
                netloc = f"{userinfo}@{netloc}"

        return urlunsplit((scheme, netloc, path, query, fragment))

    @property
    def authority(self) -> str:
        if not self.host:
            return ""
        host = self.host
        if ":" in host and not host.startswith("["):
            host = f"[{host}]"
        netloc = host
        if self.port > 0:
            netloc = f"{netloc}:{self.port}"
        if self.userinfo:
            netloc = f"{self.userinfo}@{netloc}"
        return netloc

    def join(self, ref: Union[str, "URL"]) -> "URL":
        base = self.to_string(encode=True)
        target = ref.to_string(encode=True) if isinstance(ref, URL) else ref
        return URL.parse(urljoin(base, target), normalize=True)

    # Immutable edits
    def with_scheme(self, scheme: str) -> "URL":
        scheme_n = _lower_if(scheme)
        return URL(
            scheme=scheme_n,
            userinfo=self.userinfo,
            host=self.host,
            port=_remove_default_port(scheme_n, self.host, self.port),
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

    def with_userinfo(self, userinfo: str) -> "URL":
        return URL(
            scheme=self.scheme,
            userinfo=userinfo,
            host=self.host,
            port=self.port,
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

    def with_host(self, host: str) -> "URL":
        host_n = _strip_trailing_dot(_lower_if(host))
        port_n = _remove_default_port(self.scheme, host_n, self.port)
        return URL(
            scheme=self.scheme,
            userinfo=self.userinfo,
            host=host_n,
            port=port_n,
            path=self.path,
            query=self.query,
            fragment=self.fragment,
        )

    def with_path(self, path: str) -> "URL":
        return URL(
            scheme=self.scheme,
            userinfo=self.userinfo,
            host=self.host,
            port=self.port,
            path=_normalize_path(path),
            query=self.query,
            fragment=self.fragment,
        )

    def with_query(self, query: str) -> "URL":
        return URL(
            scheme=self.scheme,
            userinfo=self.userinfo,
            host=self.host,
            port=self.port,
            path=self.path,
            query=query.lstrip("?"),
            fragment=self.fragment,
        )

    def with_fragment(self, fragment: str) -> "URL":
        return URL(
            scheme=self.scheme,
            userinfo=self.userinfo,
            host=self.host,
            port=self.port,
            path=self.path,
            query=self.query,
            fragment=fragment.lstrip("#"),
        )

    # Query helpers
    def query_items(self, *, keep_blank_values: bool = True) -> Tuple[Tuple[str, str], ...]:
        if not self.query:
            return ()
        return tuple(parse_qsl(self.query, keep_blank_values=keep_blank_values))

    def with_query_items(self, items: Mapping[str, str] | Tuple[Tuple[str, str], ...]) -> "URL":
        q = urlencode(items if isinstance(items, tuple) else list(items.items()), doseq=True)
        return self.with_query(q)

    # ------------------- URL anonymization -------------------

    def anonymize(
        self,
        mode: str = "redact",  # "redact" | "hash"
    ) -> "URL":
        """
        Returns a new URL with sensitive parts redacted.

        - Query: for keys in sensitive_query_keys, replace *values* with query_replacement.
          Keeps the key names so your logs remain groupable.
        - Userinfo: if strip_userinfo, removes user:pass@ completely.
        - Fragment: optionally redact entire fragment (sometimes people shove tokens in there).
        - No mutation; URL is frozen.

        normalize:
          - False (default): preserve existing ordering/format as much as possible, only touching
            the sensitive bits (good for debugging parity).
          - True: canonicalize like parse(normalize=True) after editing.
        """
        result = self

        if self.query:
            params = self.query_dict
            anon = anonymize_parameters(params, mode=mode)

            if params != anon:
                result = self.with_query_items(anon)

        if self.userinfo:
            result = self.with_userinfo(
                userinfo="<redacted>"
            )

        return result

    def xxh3_64(
        self,
        *,
        exclude_userinfo: bool = False,
        exclude_scheme: bool = False,
        exclude_host: bool = False,
        exclude_port: bool = False,
        exclude_path: bool = False,
        exclude_query: bool | list[str] | None = None,
        exclude_fragment: bool = False,
    ) -> int:
        """
        Stable 64-bit xxh3 hash of a canonical URL with selective exclusions.

        exclude_query semantics:
        - None           → include full query (sorted, canonical)
        - True           → exclude entire query
        - list[str]      → exclude only those query parameter names
        """
        from ..xxhash import xxhash

        scheme = "" if exclude_scheme else self.scheme
        host = "" if exclude_host else self.host
        userinfo = "" if (exclude_userinfo or exclude_host) else self.userinfo

        path = "" if exclude_path else self.path
        fragment = "" if exclude_fragment else self.fragment

        query = ""
        if exclude_query is not True:
            query = self.query
            if query:
                items = parse_qsl(query, keep_blank_values=True)
                if isinstance(exclude_query, list):
                    excl = set(exclude_query)
                    items = [(k, v) for (k, v) in items if k not in excl]
                items.sort(key=lambda kv: (kv[0], kv[1]))
                query = urlencode(items, doseq=True)

        port = _NO_PORT if (exclude_port or exclude_host) else self.port

        netloc = ""
        if host:
            out_host = host
            if ":" in out_host and not out_host.startswith("["):
                out_host = f"[{out_host}]"
            netloc = out_host

            if port > 0:
                netloc = f"{netloc}:{port}"

            if userinfo:
                netloc = f"{userinfo}@{netloc}"

        s = urlunsplit((scheme, netloc, path, query, fragment))

        h = xxhash.xxh3_64()
        h.update(s.encode("utf-8"))
        return h



from __future__ import annotations

import json as json_module
from dataclasses import dataclass, replace
from typing import Mapping, Any, Optional, MutableMapping

import pyarrow as pa
from yggdrasil.io.headers import anonymize_headers

from .dynamic_buffer import DynamicBuffer
from .url import URL

__all__ = ["PreparedRequest"]


@dataclass
class PreparedRequest:
    method: str
    url: URL
    headers: Mapping[str, str]
    buffer: Optional[DynamicBuffer]
    sent_at_timestamp: int = 0  # time.time_ns() // 1000

    @classmethod
    def prepare(
        cls,
        method: str,
        url: URL | str,
        headers: Optional[MutableMapping[str, str]] = None,
        body: Optional[Any] = None,
        *,
        json: Optional[Any] = None,
        normalize: bool = True,
    ) -> "PreparedRequest":
        url = URL.parse_any(url, normalize=normalize)

        if body is not None:
            body = DynamicBuffer.parse_any(obj=body)
        elif json is not None:
            body = DynamicBuffer()
            json_module.dump(json, body)
            body.seek(0)

            if not headers:
                headers = {"Content-Type": "application/json"}
            else:
                headers["Content-Type"] = "application/json"

        if not headers:
            headers = {}

        if body is not None:
            headers["Content-Length"] = body.size

        return cls(method=method, url=url, headers=headers, buffer=body, sent_at_timestamp=0)

    @classmethod
    def parse_any(
        cls,
        obj: Any,
        *,
        normalize: bool = True,
    ) -> "PreparedRequest":
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, str):
            return cls.parse_str(obj, normalize=normalize)

        if isinstance(obj, Mapping):
            return cls.parse_dict(obj, normalize=normalize)

        # last-resort: stringify (useful for weird wrappers)
        return cls.parse_str(str(obj), normalize=normalize)

    @classmethod
    def parse_str(
        cls,
        raw: str,
        *,
        normalize: bool = True,
    ) -> "PreparedRequest":
        """
        Accepts JSON string representing a PreparedRequest-ish dict.
        """
        s = raw.strip()
        if not s:
            raise ValueError("PreparedRequest.parse_str: empty string")

        try:
            d = json_module.loads(s)
        except Exception as e:
            raise ValueError("PreparedRequest.parse_str: expected JSON object string") from e

        if not isinstance(d, Mapping):
            raise ValueError("PreparedRequest.parse_str: JSON must decode to an object")

        return cls.parse_dict(d, normalize=normalize)

    @classmethod
    def parse_dict(
        cls,
        d: Mapping[str, Any],
        *,
        normalize: bool = True,
    ) -> "PreparedRequest":
        """
        Parses a PreparedRequest from a mapping.

        Supported shapes (field aliases included):
          - method: "method" | "http_method" | "verb"
          - url:    "url" | "uri" | "href" | ("url_*" exploded fields)
          - headers:"headers" | "header" | "hdrs"
          - body:   "buffer" | "body" | "data"
          - sent_at:"sent_at_timestamp" | "sent_at" | "timestamp" | "time_ns"
        """
        if not d:
            raise ValueError("PreparedRequest.parse_dict: empty mapping")

        # method
        method = (
            d.get("method")
            or d.get("http_method")
            or d.get("verb")
            or "GET"
        )
        method_s = str(method)

        # url (accept URL object, string, dict, or exploded url_* fields)
        url_obj: Any = d.get("url") or d.get("uri") or d.get("href")
        if url_obj is None:
            # exploded fields as used by to_arrow_batch
            has_exploded = any(
                k in d
                for k in (
                    "url_scheme",
                    "url_userinfo",
                    "url_host",
                    "url_port",
                    "url_path",
                    "url_query",
                    "url_fragment",
                )
            )
            if has_exploded:
                q = d.get("url_query")
                query = ""
                if isinstance(q, Mapping):
                    # our arrow encoding is key -> "v1|v2|..."
                    # keep deterministic: sort keys, keep given value order
                    parts: list[tuple[str, str]] = []
                    for k in sorted(q.keys(), key=lambda x: str(x)):
                        v = q[k]
                        if v is None:
                            continue
                        vs = str(v).split("|")
                        for one in vs:
                            parts.append((str(k), one))
                    from urllib.parse import urlencode
                    query = urlencode(parts, doseq=True)

                url_obj = {
                    "scheme": d.get("url_scheme") or "",
                    "userinfo": d.get("url_userinfo") or "",
                    "host": d.get("url_host") or "",
                    "port": d.get("url_port") or 0,
                    "path": d.get("url_path") or "",
                    "query": query,
                    "fragment": d.get("url_fragment") or "",
                }
            else:
                raise ValueError("PreparedRequest.parse_dict: missing url")

        url = URL.parse_any(url_obj, normalize=normalize)

        # headers
        headers_obj = d.get("headers") or d.get("header") or d.get("hdrs") or {}
        if headers_obj is None:
            headers_obj = {}
        if not isinstance(headers_obj, Mapping):
            raise ValueError("PreparedRequest.parse_dict: headers must be a mapping")
        headers: dict[str, str] = {str(k): str(v) for k, v in headers_obj.items()}

        # body/buffer
        body_obj = d.get("buffer")
        if body_obj is None:
            body_obj = d.get("body")
        if body_obj is None:
            body_obj = d.get("data")

        buffer: Optional[DynamicBuffer] = None
        if body_obj is not None:
            buffer = DynamicBuffer.parse_any(obj=body_obj)

        # sent_at timestamp (ns)
        sent_at = (
            d.get("sent_at_timestamp")
            if "sent_at_timestamp" in d
            else d.get("time_ns")
            if "time_ns" in d
            else d.get("timestamp")
            if "timestamp" in d
            else d.get("sent_at")
        )
        sent_at_ts = 0
        if sent_at is not None and sent_at != "":
            if isinstance(sent_at, int):
                sent_at_ts = sent_at
            else:
                s = str(sent_at)
                sent_at_ts = int(s) if s.isdigit() else 0

        return cls(
            method=method_s,
            url=url,
            headers=headers,
            buffer=buffer,
            sent_at_timestamp=sent_at_ts,
        )

    @property
    def body(self):
        return self.buffer

    def anonymize(
        self,
        mode: str = "redact"
    ) -> "PreparedRequest":
        """
        Clean/boring + composable:
        - headers redaction happens here
        - URL redaction happens in URL.anonymize()
        """
        return replace(
            self,
            headers=anonymize_headers(self.headers, mode=mode),
            url=self.url.anonymize(mode=mode),
        )

    def to_arrow_batch(
        self,
        parse: bool = False,
        *,
        column_prefix: str = "",
    ) -> pa.RecordBatch:
        if parse:
            raise NotImplementedError

        schema = pa.schema(
            [
                pa.field(f"{column_prefix}method", pa.string(), nullable=False, metadata={"comment": "The HTTP verb (GET, POST, etc.)"}),
                pa.field(f"{column_prefix}url", pa.string(), nullable=False, metadata={"comment": "The full request URL string"}),
                pa.field(f"{column_prefix}url_scheme", pa.string(), nullable=True, metadata={"comment": "URL protocol (e.g., http, https)"}),
                pa.field(f"{column_prefix}url_userinfo", pa.string(), nullable=True, metadata={"comment": "Authentication information in the URL"}),
                pa.field(f"{column_prefix}url_host", pa.string(), nullable=True, metadata={"comment": "Domain name or IP address of the server"}),
                pa.field(f"{column_prefix}url_port", pa.int32(), nullable=True, metadata={"comment": "TCP port number"}),
                pa.field(f"{column_prefix}url_path", pa.string(), nullable=True, metadata={"comment": "Hierarchical path to the resource"}),
                pa.field(
                    f"{column_prefix}url_query",
                    pa.map_(pa.field("key", pa.string(), nullable=False), pa.field("value", pa.string(), nullable=False)),
                    nullable=True,
                    metadata={"comment": "Parsed query string parameters as key-value pairs"},
                ),
                pa.field(f"{column_prefix}url_fragment", pa.string(), nullable=True, metadata={"comment": "The internal anchor or fragment identifier"}),
                pa.field(f"{column_prefix}body", pa.binary(), nullable=True, metadata={"comment": "Raw binary payload of the request"}),
                pa.field(f"{column_prefix}body_hash64", pa.int64(), nullable=True, metadata={"comment": "64-bit hash (xxh3) of the request body", "algorithm": "xxh3_64"}),
                pa.field(f"{column_prefix}sent_at", pa.timestamp("us", "UTC"), nullable=False, metadata={"comment": "UTC timestamp of when the request was dispatched"}),
            ]
        )

        u = self.url
        url_s = u.to_string()

        scheme_v = u.scheme or None
        userinfo_v = u.userinfo or None
        host_v = u.host or None
        port_v = None if (u.port is None or u.port == 0) else int(u.port)
        path_v = u.path or None
        fragment_v = u.fragment or None

        q = u.query_dict
        q_v = None if not q else {k: "|".join(vs) for k, vs in q.items()}

        body_bytes, body_h64 = (None, None) if self.buffer is None else (self.buffer.to_bytes(), self.buffer.xxh3_64().intdigest())

        arrays = [
            pa.array([self.method], type=pa.string()),
            pa.array([url_s], type=pa.string()),
            pa.array([scheme_v], type=pa.string()),
            pa.array([userinfo_v], type=pa.string()),
            pa.array([host_v], type=pa.string()),
            pa.array([port_v], type=pa.int32()),
            pa.array([path_v], type=pa.string()),
            pa.array([q_v], type=schema.field(f"{column_prefix}url_query").type),
            pa.array([fragment_v], type=pa.string()),
            pa.array([body_bytes], type=pa.binary()),
            pa.array([body_h64], type=pa.int64()),
            pa.array([self.sent_at_timestamp], type=pa.timestamp("us", "UTC")),
        ]

        return pa.RecordBatch.from_arrays(arrays, schema=schema)  # type: ignore



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterator, Literal, Optional, Union

import pyarrow as pa
from yggdrasil.pyutils.waiting_config import WaitingConfig, WaitingConfigArg

if TYPE_CHECKING:
    import pandas
    import polars
    import pyarrow.dataset as ds
    import pyspark.sql


__all__ = ["StatementResult"]


@dataclass
class StatementResult(ABC):
    """Arrow-first result wrapper with fast, ergonomic conversions.

    This base class models a "statement-like" execution result whose best interchange
    format is Apache Arrow (RecordBatches / Tables).

    Implementations MUST provide:
    - `done`: terminal-state indicator
    - `failed`: failure/cancellation indicator
    - `raise_for_status()`: raise an exception on failure/cancellation (no-op on success)
    - `refresh_status()`: update status from the backend
    - `arrow_schema`: result schema
    - `to_arrow_reader()`: produce a `pyarrow.RecordBatchReader` over result batches

    Everything else (batches / Dataset / Table / pandas / Polars / Spark) is built on top.

    Notes
    -----
    - This class maintains optional, internal caches: `_arrow_table` and `_spark_df`.
      They are NOT dataclass fields on purpose, to avoid them appearing in repr/equality,
      and to make subclass dataclass layouts frictionless.
    - If `_spark_df` is present, conversions that go through Arrow (`toArrow()`) collect
      to the driver.
    - `stream` in this API means "prefer streaming / lazy execution" for consumers that
      support it (notably Polars). For Arrow itself, streaming behavior is determined by
      the concrete implementation of `to_arrow_reader()`.
    """

    # -------------------------------------------------------------------------
    # Lifecycle / internal caches (NOT dataclass fields)
    # -------------------------------------------------------------------------

    def __post_init__(self) -> None:
        # These are intentionally not dataclass fields.
        self._arrow_table: Optional[pa.Table] = None
        self._spark_df: Optional["pyspark.sql.DataFrame"] = None

    def __iter__(self) -> Iterator[pa.RecordBatch]:
        """Iterate over Arrow record batches.

        Equivalent to calling `to_arrow_batches()` with default parameters.
        """
        return self.to_arrow_batches()

    # -------------------------------------------------------------------------
    # Pickling
    # -------------------------------------------------------------------------

    def __getstate__(self) -> dict:
        """Prepare pickle-ready state.

        Spark DataFrames are not pickleable. If `_spark_df` exists, it is converted to a
        local Arrow table and stored into `_arrow_table` in the pickled state.

        Returns
        -------
        dict
            Copy of instance state with Spark payload converted to Arrow when needed.

        Notes
        -----
        This conversion collects to the driver, so do not pickle instances holding huge
        Spark DataFrames unless that's explicitly intended.
        """
        state = self.__dict__.copy()
        spark_df = state.pop("_spark_df", None)

        if spark_df is not None:
            from ..spark.cast import spark_dataframe_to_arrow_table
            state["_arrow_table"] = spark_dataframe_to_arrow_table(spark_df, None)

        # Ensure stable unpickle shape.
        state.setdefault("_arrow_table", None)
        state.setdefault("_spark_df", None)
        return state

    def __setstate__(self, state: dict) -> None:
        """Restore instance state from pickle."""
        self.__dict__.update(state)
        # Spark DF cannot be restored; keep it None unless caller manually reattaches.
        self.__dict__.setdefault("_arrow_table", None)
        self.__dict__.setdefault("_spark_df", None)

    # -------------------------------------------------------------------------
    # Core state / caching
    # -------------------------------------------------------------------------

    @property
    def is_spark_sql(self) -> bool:
        """True if this result currently has a Spark DataFrame attached."""
        return self._spark_df is not None

    @property
    def persisted(self) -> bool:
        """Whether results are already materialized locally."""
        return self._arrow_table is not None or self._spark_df is not None

    def persist(self, mode: Literal["arrow", "spark"] = "arrow") -> "StatementResult":
        """Materialize and cache this result for cheaper subsequent conversions.

        Parameters
        ----------
        mode : {"arrow", "spark"}
            - "arrow": materialize to a local `pyarrow.Table` and store in `_arrow_table`.
            - "spark": materialize to a Spark DataFrame and store in `_spark_df`.

        Returns
        -------
        StatementResult
            Self (mutated in-place by setting `_arrow_table` or `_spark_df`).

        Notes
        -----
        Persisting to Spark may still collect to the driver first if the only available
        representation is Arrow.
        """
        if self.persisted:
            return self

        if mode == "arrow":
            self._arrow_table = self.to_arrow_table()
            return self

        if mode == "spark":
            self._spark_df = self.to_spark(prefer_cached=False, cache_result=False)
            return self

        raise ValueError(f"Unknown persist mode: {mode!r}. Expected 'arrow' or 'spark'.")

    # -------------------------------------------------------------------------
    # Execution lifecycle contract
    # -------------------------------------------------------------------------

    @property
    @abstractmethod
    def done(self) -> bool:
        """True when the statement is in a terminal state."""
        raise NotImplementedError

    @property
    @abstractmethod
    def failed(self) -> bool:
        """True when the statement failed or was cancelled."""
        raise NotImplementedError

    @abstractmethod
    def raise_for_status(self) -> None:
        """Raise an exception if the statement failed/cancelled; no-op on success."""
        raise NotImplementedError

    @abstractmethod
    def refresh_status(self) -> None:
        """Refresh the statement status from the backend."""
        raise NotImplementedError

    def wait(self, wait: WaitingConfigArg = True, raise_error: bool = True) -> "StatementResult":
        """Block until the statement reaches a terminal state or timeout.

        Parameters
        ----------
        wait : WaitingConfigArg
            Waiting configuration.
            - True: use default waiting behavior (poll + timeout from defaults)
            - False: do not wait
            - WaitingConfig: custom waiting behavior (sleep/backoff/timeout)
        raise_error : bool
            If True, raise after reaching terminal state if failed/cancelled.

        Returns
        -------
        StatementResult
            Self.
        """
        wait_cfg = WaitingConfig.check_arg(wait)

        # Convention: falsy timeout => don't wait (or waiting disabled)
        if not wait_cfg.timeout:
            if raise_error:
                self.raise_for_status()
            return self

        iteration, start = 0, time.time()

        while not self.done:
            wait_cfg.sleep(iteration=iteration, start=start)
            iteration += 1
            self.refresh_status()

        if raise_error:
            self.raise_for_status()

        return self

    # -------------------------------------------------------------------------
    # Arrow contract
    # -------------------------------------------------------------------------

    @property
    @abstractmethod
    def arrow_schema(self) -> pa.Schema:
        """Arrow schema for the result set."""
        raise NotImplementedError

    @abstractmethod
    def to_arrow_reader(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> pa.RecordBatchReader:
        """Return a streaming Arrow `RecordBatchReader` over the results.

        Parameters
        ----------
        max_workers : int
            Maximum parallel workers used for remote chunk downloads / decoding.
        max_in_flight : Optional[int]
            Backpressure limit for concurrently in-flight remote requests/batches.
        batch_size : Optional[int]
            Target batch size for client-side batching/coalescing. Hint only.
        schema : Optional[pyarrow.Schema]
            Override schema. Defaults to `self.arrow_schema`.
        maintain_order : bool
            Preserve original ordering of remote chunks/batches when streaming.
        stream : bool
            Prefer a streaming reader implementation (if the backend supports it).

        Returns
        -------
        pyarrow.RecordBatchReader
        """
        raise NotImplementedError

    # -------------------------------------------------------------------------
    # Arrow / Dataset conversions
    # -------------------------------------------------------------------------

    def to_arrow_batches(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> Iterator[pa.RecordBatch]:
        """Iterate over result batches as `pyarrow.RecordBatch`."""
        resolved_schema = schema or self.arrow_schema

        reader = self.to_arrow_reader(
            max_workers=max_workers,
            max_in_flight=max_in_flight,
            batch_size=batch_size,
            schema=resolved_schema,
            maintain_order=maintain_order,
            stream=stream,
        )

        yield from reader

    def to_arrow_dataset(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> "ds.Dataset":
        """Return an in-memory `pyarrow.dataset.Dataset` view of the results.

        Best bridge into lazy engines, especially Polars:
        `polars.scan_pyarrow_dataset(result.to_arrow_dataset(...))`.

        Notes
        -----
        If backed by Spark, `toArrow()` collects to the driver.
        """
        import pyarrow.dataset as pds

        resolved_schema = schema or self.arrow_schema

        if self.persisted:
            if self._arrow_table is not None:
                return pds.dataset(self._arrow_table, schema=resolved_schema)

            if self._spark_df is not None:
                tbl = self._spark_df.toArrow()
                batches = tbl.to_batches(max_chunksize=batch_size) if batch_size else tbl.to_batches()
                return pds.dataset(batches, schema=resolved_schema)

            raise NotImplementedError("Persisted without Arrow table or Spark DF")

        reader = self.to_arrow_reader(
            max_workers=max_workers,
            max_in_flight=max_in_flight,
            batch_size=batch_size,
            schema=resolved_schema,
            maintain_order=maintain_order,
            stream=stream,
        )
        return pds.dataset(reader)

    def to_arrow_table(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> pa.Table:
        """Materialize results into a `pyarrow.Table` (collects all batches)."""
        if self.persisted:
            if self._arrow_table is not None:
                return self._arrow_table
            if self._spark_df is not None:
                return self._spark_df.toArrow()
            raise NotImplementedError("Persisted without Arrow table or Spark DF")

        resolved_schema = schema or self.arrow_schema

        reader = self.to_arrow_reader(
            max_workers=max_workers,
            max_in_flight=max_in_flight,
            batch_size=batch_size,
            schema=resolved_schema,
            maintain_order=maintain_order,
            stream=stream,
        )

        return reader.read_all()

    # -------------------------------------------------------------------------
    # pandas / polars
    # -------------------------------------------------------------------------

    def to_pandas(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> "pandas.DataFrame":
        """Materialize into a pandas DataFrame via Arrow."""
        return self.to_arrow_table(
            max_workers=max_workers,
            max_in_flight=max_in_flight,
            batch_size=batch_size,
            schema=schema,
            maintain_order=maintain_order,
            stream=stream,
        ).to_pandas()

    def to_polars(
        self,
        *,
        max_workers: int = 4,
        max_in_flight: Optional[int] = None,
        batch_size: Optional[int] = None,
        schema: Optional[pa.Schema] = None,
        maintain_order: bool = False,
        stream: bool = True,
    ) -> Union["polars.LazyFrame", "polars.DataFrame"]:
        """Convert results to Polars.

        - stream=True  -> returns `polars.LazyFrame` scanning an Arrow Dataset
        - stream=False -> returns `polars.DataFrame` by collecting immediately
        """
        from ..polars.lib import polars

        arrow_dataset = self.to_arrow_dataset(
            max_workers=max_workers,
            max_in_flight=max_in_flight,
            batch_size=batch_size,
            schema=schema,
            maintain_order=maintain_order,
            stream=stream,
        )

        lf = polars.scan_pyarrow_dataset(arrow_dataset)
        return lf if stream else lf.collect()

    # -------------------------------------------------------------------------
    # Spark
    # -------------------------------------------------------------------------

    def to_spark(
        self,
        *,
        spark: Optional["pyspark.sql.SparkSession"] = None,
        prefer_cached: bool = True,
        cache_result: bool = True,
    ) -> "pyspark.sql.DataFrame":
        """Convert results to a Spark DataFrame.

        Notes
        -----
        If `_spark_df` is not present, this materializes an Arrow table on the driver and
        then creates a Spark DataFrame. This is driver-memory-bound.
        """
        if prefer_cached and self._spark_df is not None:
            return self._spark_df

        from ..spark.cast import arrow_table_to_spark_dataframe

        sdf = arrow_table_to_spark_dataframe(self.to_arrow_table(), spark)

        if cache_result:
            self._spark_df = sdf

        return sdf



# yggdrasil/pyutils/dynamic_buffer.py
from __future__ import annotations

import io
import os
import struct
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union, IO, Any, TYPE_CHECKING


__all__ = ["DynamicBuffer", "DynamicBufferConfig"]


if TYPE_CHECKING:
    import xxhash


@dataclass(frozen=False, slots=True)
class DynamicBufferConfig:
    """
    Configuration for DynamicBuffer.

    spill_bytes:
        When buffered data size exceeds this threshold, switch from in-memory BytesIO
        to a real temporary file on disk.

    tmp_dir:
        Directory where spill files are created. If None, uses system temp dir.

    prefix/suffix:
        Naming for the temp file when spilling to disk.

    keep_spilled_file:
        If True, spilled temp files are not deleted on close (you can call cleanup()).
        If False, we delete the temp file on close/cleanup().
    """
    spill_bytes: int = 128 * 1024 * 1024  # 64 MiB
    tmp_dir: Optional[Path] = None
    prefix: str = "tmp-"
    suffix: str = ".bin"
    keep_spilled_file: bool = False


class DynamicBuffer(io.RawIOBase):
    """
    A bytes buffer that starts in memory and spills to a local temp file
    when it grows beyond a threshold.

    Practical uses:
      - serialization buffers (dill/pickle/parquet writes)
      - staging upload payloads without blowing RAM
      - "read as bytes" semantics while supporting large payloads

    Key properties:
      - file-like: supports write(), read(), seek(), tell(), flush(), close()
      - getvalue() works only if still in-memory (else raises)
      - to_bytes() returns bytes (loads entire payload; be careful for huge buffers)
      - path returns the spill file path if spilled, else None
      - underlying IO always positioned at end after write()

    Notes:
      - This is a byte buffer (binary mode).
      - Not thread-safe (like most IO objects).
    """

    def __init__(
        self,
        config: DynamicBufferConfig | None = None
    ) -> None:
        super().__init__()
        self._cfg = config or DynamicBufferConfig()
        if self._cfg.spill_bytes < 1:
            raise ValueError("spill_bytes must be >= 1")

        self._mem: io.BytesIO | None = io.BytesIO()
        self._file: IO[bytes] | None = None
        self._path: Path | None = None
        self._closed: bool = False

    def __bytes__(self) -> bytes:
        return self.to_bytes()

    def __len__(self) -> int:
        """
        Total payload size in bytes (does NOT change the current cursor).
        """
        return self.size

    def __del__(self):
        self.close()

    @classmethod
    def parse_any(
        cls,
        obj: Any,
        config: DynamicBufferConfig | None = None
    ):
        if isinstance(obj, cls):
            return obj

        buffer = DynamicBuffer(config=config)

        if isinstance(obj, io.BytesIO):
            obj._mem = obj
        elif isinstance(obj, (str, Path)):
            from .path import LocalDataPath

            obj._path = LocalDataPath(obj)
            obj._file = obj._path.open("w+b", buffering=0)
        else:
            buffer.write_any_bytes(obj)
            buffer.seek(0)

        return buffer

    # -----------------------
    # Introspection
    # -----------------------
    @property
    def spilled(self) -> bool:
        return self._file is not None

    @property
    def path(self) -> Path | None:
        return self._path

    @property
    def size(self) -> int:
        if self._mem is not None:
            # BytesIO: safest is buffer length; avoids messing with cursor.
            return self._mem.getbuffer().nbytes

        try:
            return os.fstat(self._file.fileno()).st_size
        except Exception:
            # fallback: preserve cursor, use seek/tell
            pos = self._file.tell()
            self._file.seek(0, io.SEEK_END)
            end = self._file.tell()
            self._file.seek(pos, io.SEEK_SET)
            return int(end)

    # -----------------------
    # io.RawIOBase interface
    # -----------------------
    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def close(self) -> None:
        if self._closed:
            return
        try:
            self.flush()
        except Exception:
            pass

        try:
            if self._mem is not None:
                self._mem.close()
        finally:
            self._mem = None

        try:
            if self._file is not None:
                self._file.close()
        finally:
            self._file = None

        if self._path is not None and not self._cfg.keep_spilled_file:
            try:
                self._path.unlink(missing_ok=True)  # py3.8+ ok
            except Exception:
                pass

        self._closed = True
        super().close()

    def flush(self) -> None:
        fh = self._fh()
        try:
            fh.flush()
        except Exception:
            pass

    def tell(self) -> int:
        return self._fh().tell()

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self._fh().seek(offset, whence)

    def read(self, size: int = -1) -> bytes:
        return self._fh().read(size)

    def write(self, b: Union[bytes, bytearray, memoryview]) -> int:
        if self._closed:
            raise ValueError("I/O operation on closed DynamicBuffer")

        try:
            mv = memoryview(b)
        except TypeError:
            if isinstance(b, str):
                mv = memoryview(b.encode("utf-8"))
            else:
                raise

        n = len(mv)

        # If we're still in memory, check whether this write pushes us over.
        if self._mem is not None:
            current_size = self._mem.getbuffer().nbytes
            if current_size + n > self._cfg.spill_bytes:
                self._spill_to_file()

        fh = self._fh()
        written = fh.write(mv.tobytes())
        return written

    def write_any_bytes(self, obj: Any):
        if obj is None:
            return 0

        if isinstance(obj, (bytes, bytearray, memoryview)):
            return self.write(obj)

        blob = obj.read()

        return self.write(blob)

    # -----------------------
    # Structured binary read/write (little-endian)
    # -----------------------

    def _read_exact(self, n: int) -> bytes:
        data = self.read(n)
        if len(data) != n:
            raise EOFError(f"expected {n} bytes, got {len(data)}")
        return data

    # ---- ints ----

    def read_int8(self) -> int:
        return struct.unpack("<b", self._read_exact(1))[0]

    def write_int8(self, value: int) -> int:
        return self.write(struct.pack("<b", int(value)))

    def read_uint8(self) -> int:
        return struct.unpack("<B", self._read_exact(1))[0]

    def write_uint8(self, value: int) -> int:
        return self.write(struct.pack("<B", int(value)))

    def read_int16(self) -> int:
        return struct.unpack("<h", self._read_exact(2))[0]

    def write_int16(self, value: int) -> int:
        return self.write(struct.pack("<h", int(value)))

    def read_uint16(self) -> int:
        return struct.unpack("<H", self._read_exact(2))[0]

    def write_uint16(self, value: int) -> int:
        return self.write(struct.pack("<H", int(value)))

    def read_int32(self) -> int:
        return struct.unpack("<i", self._read_exact(4))[0]

    def write_int32(self, value: int) -> int:
        return self.write(struct.pack("<i", int(value)))

    def read_uint32(self) -> int:
        return struct.unpack("<I", self._read_exact(4))[0]

    def write_uint32(self, value: int) -> int:
        return self.write(struct.pack("<I", int(value)))

    def read_int64(self) -> int:
        return struct.unpack("<q", self._read_exact(8))[0]

    def write_int64(self, value: int) -> int:
        return self.write(struct.pack("<q", int(value)))

    def read_uint64(self) -> int:
        return struct.unpack("<Q", self._read_exact(8))[0]

    def write_uint64(self, value: int) -> int:
        return self.write(struct.pack("<Q", int(value)))

    # ---- floats ----

    def read_f32(self) -> float:
        return struct.unpack("<f", self._read_exact(4))[0]

    def write_f32(self, value: float) -> int:
        return self.write(struct.pack("<f", float(value)))

    def read_f64(self) -> float:
        return struct.unpack("<d", self._read_exact(8))[0]

    def write_f64(self, value: float) -> int:
        return self.write(struct.pack("<d", float(value)))

    # ---- bool ----

    def read_bool(self) -> bool:
        return bool(self.read_uint8())

    def write_bool(self, value: bool) -> int:
        return self.write_uint8(1 if value else 0)

    # ---- bytes / strings (length-prefixed) ----

    def read_bytes_u32(self) -> bytes:
        """
        Read: u32 length (little-endian) + payload bytes.
        """
        n = self.read_uint32()
        return self._read_exact(n)

    def write_bytes_u32(self, data: bytes | bytearray | memoryview) -> int:
        mv = memoryview(data)
        total = self.write_uint32(len(mv))
        total += self.write(mv)
        return total

    def read_str_u32(self, encoding: str = "utf-8") -> str:
        return self.read_bytes_u32().decode(encoding)

    def write_str_u32(self, s: str, encoding: str = "utf-8") -> int:
        b = s.encode(encoding)
        return self.write_bytes_u32(b)

    # -----------------------
    # Convenience APIs
    # -----------------------
    def xxh3_64(self) -> "xxhash.xxh3_64":
        from ..xxhash import xxhash

        h = xxhash.xxh3_64()
        h.update(self.to_bytes())
        return h

    def getvalue(self) -> bytes:
        """
        Return bytes only if still in memory.
        Raises if spilled to disk (to avoid accidental huge loads).
        """
        if self._mem is None:
            return self.to_bytes()
        return self._mem.getvalue()

    def to_bytes(self) -> bytes:
        """
        Read the entire buffer into memory.
        Be careful: if you spilled, this could be huge.
        """
        fh = self._fh()
        pos = fh.tell()
        try:
            fh.seek(0)
            return fh.read()
        finally:
            try:
                fh.seek(pos)
            except Exception:
                pass

    def open_reader(self) -> IO[bytes]:
        """
        Return a new readable handle positioned at start.
        - If in-memory: returns a new BytesIO copy
        - If spilled: opens the spill file in 'rb'
        """
        if self._mem is not None:
            return io.BytesIO(self._mem.getvalue())
        if self._path is None:
            raise RuntimeError("Spilled buffer missing path (unexpected)")
        return open(self._path, "rb")

    def open_writer(self) -> IO[bytes]:
        """
        Return the underlying writable handle (position preserved).
        Mostly for advanced scenarios.
        """
        return self._fh()

    def cleanup(self) -> None:
        """
        Explicit cleanup (same behavior as close regarding file deletion).
        Safe to call multiple times.
        """
        self.close()

    # -----------------------
    # Internals
    # -----------------------
    def _fh(self) -> IO[bytes]:
        if self._closed:
            raise ValueError("I/O operation on closed DynamicBuffer")
        if self._file is not None:
            return self._file
        if self._mem is not None:
            return self._mem
        raise RuntimeError("DynamicBuffer in invalid state")

    def _spill_to_file(self) -> None:
        """
        Create a temp file, copy existing mem contents, and switch to disk.
        Preserves current cursor position.
        """
        if self._mem is None:
            return

        cfg = self._cfg

        tmp_dir = str(cfg.tmp_dir) if cfg.tmp_dir is not None else None
        fd, name = tempfile.mkstemp(prefix=cfg.prefix, suffix=cfg.suffix, dir=tmp_dir)
        path = Path(name)

        # We want a binary file handle with random access.
        # Use os.fdopen so the fd is managed correctly.
        f = os.fdopen(fd, "w+b", buffering=0)

        # Copy bytes + preserve cursor position
        mem = self._mem
        mem_pos = mem.tell()
        mem.seek(0)
        f.write(mem.read())
        f.flush()
        f.seek(mem_pos)

        # Switch backing store
        mem.close()
        self._mem = None
        self._file = f
        self._path = path



"""Parallel execution decorator utilities."""

from __future__ import annotations

import concurrent.futures as cf
from contextlib import nullcontext
from functools import wraps
from multiprocessing.reduction import ForkingPickler
from typing import (
    Callable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    ParamSpec,
)

import dill

ForkingPickler.loads = dill.loads
ForkingPickler.dumps = dill.dumps

P = ParamSpec("P")
R = TypeVar("R")


def parallelize(
    executor_cls: Type[cf.Executor] = cf.ThreadPoolExecutor,
    *,
    max_workers: Optional[int] = None,
    arg_index: int = 0,
    timeout: Optional[float] = None,
    return_exceptions: bool = False,
    show_progress: bool = False,  # 👈 new flag
) -> Callable[[Callable[P, R]], Callable[P, Iterator[R]]]:
    """
    Decorator to parallelize a function/method over one iterable argument
    using a concurrent.futures.Executor.

    Returns
    -------
    A wrapper that returns an iterator (generator) of results, not a list.
    """

    def decorator(func: Callable[P, R]) -> Callable[P, Iterator[R]]:
        """Wrap a callable to execute over an iterable using an executor.

        Args:
            func: Callable to wrap.

        Returns:
            Wrapped callable that yields results.
        """
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterator[R]:  # type: ignore[misc]
            """Execute the wrapped function across items in parallel.

            Args:
                *args: Positional args for the wrapped function.
                **kwargs: Keyword args for the wrapped function.

            Returns:
                Iterator of results.
            """
            # Basic sanity checks
            if arg_index < 0 or arg_index >= len(args):
                raise ValueError(
                    f"arg_index {arg_index} out of range for {len(args)} positional args"
                )

            iterable_arg = args[arg_index]

            # Try to know total upfront for proper progress bar
            try:
                total = len(iterable_arg)  # type: ignore[arg-type]
            except TypeError:
                total = None

            try:
                iterator = iter(iterable_arg)  # type: ignore[arg-type]
            except TypeError:
                raise TypeError(
                    f"Argument at position {arg_index} must be iterable "
                    f"for parallel execution, got {type(iterable_arg)!r}"
                )

            # Split args into prefix / iterable / suffix so we can rebuild per task
            prefix = args[:arg_index]
            suffix = args[arg_index + 1 :]

            # Allow caller to provide existing executor via kwarg
            executor: Optional[cf.Executor] = kwargs.pop("executor", None)
            owns_executor = executor is None

            if executor is None:
                executor = executor_cls(max_workers=max_workers)

            # Generator that will actually submit tasks and yield results
            def gen() -> Iterator[R]:
                """Yield results from parallel execution in input order.

                Yields:
                    Results in input order.
                """
                futures: list[cf.Future[R]] = []
                processed = 0  # for progress

                # If we created the executor, manage its lifetime with a context manager.
                # If executor is external, use nullcontext so we don't close it.
                ctx = executor if owns_executor else nullcontext(executor)

                def _print_progress() -> None:
                    """Emit progress output when enabled.

                    Returns:
                        None.
                    """
                    if not show_progress:
                        return

                    nonlocal processed
                    processed += 1

                    # Known total: real progress bar
                    if total is not None and total > 0:
                        width = 40
                        frac = processed / total
                        filled = int(width * frac)
                        bar = "#" * filled + "-" * (width - filled)
                        print(
                            f"\r[{bar}] {processed}/{total} ({frac:6.1%})",
                            end="",
                            flush=True,
                        )
                        if processed == total:
                            print()  # newline at end
                    else:
                        # Unknown total: simple counter
                        print(f"\rProcessed {processed} items", end="", flush=True)

                with ctx:
                    # Submit all tasks first so they can run in parallel
                    for item in iterator:
                        call_args = (*prefix, item, *suffix)
                        futures.append(
                            executor.submit(  # type: ignore[arg-type]
                                func,
                                *call_args,
                                **kwargs,
                            )
                        )

                    # Yield results in input order
                    for fut in futures:
                        try:
                            res = fut.result(timeout=timeout)
                        except Exception as e:
                            if return_exceptions:
                                _print_progress()
                                # type: ignore[list-item]
                                yield e  # type: ignore[misc]
                                continue
                            # Best-effort cancel of remaining futures if we own the executor
                            if owns_executor:
                                for f2 in futures:
                                    if not f2.done():
                                        f2.cancel()
                            # ensure progress line ends cleanly
                            if show_progress:
                                print()
                            raise
                        else:
                            _print_progress()
                            yield res

                # if we printed a counter with unknown total, finish with newline
                if show_progress and (total is None or total == 0):
                    print()

            # Return generator; execution happens when iterated
            return gen()

        return wrapper

    return decorator



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Never, Iterable, Optional
import importlib

__all__ = ["Dummy"]


@dataclass(frozen=True, slots=True)
class Dummy:
    _module_path: tuple[str, ...]

    @classmethod
    def from_name(
        cls,
        *module_path: str | Iterable[str],
        to_class: Optional[bool] = None,
        try_import: bool = False,
    ) -> Any:
        """
        Examples:
          Dummy.from_name("pyarrow", try_import=True, to_class=False) -> module or Dummy
          Dummy.from_name("pyarrow", "compute", to_class=False)       -> Dummy(("pyarrow","compute"))
          Dummy.from_name(["pyarrow","compute"], to_class=False)      -> Dummy(("pyarrow","compute"))

          Dummy.from_name("pyarrow", "Table", to_class=True, try_import=True) -> pa.Table (class) or Dummy
          Dummy.from_name("pyarrow", "compute", "cast", to_class=True, try_import=True) -> resolved attr or Dummy
        """
        # Normalize inputs into a flat tuple[str, ...]
        parts: list[str] = []
        for p in module_path:
            if isinstance(p, str):
                parts.append(p)
            elif isinstance(p, Iterable):
                parts.extend(str(x) for x in p)
            else:
                parts.append(str(p))

        path = tuple(parts) if parts else ("<unknown>",)

        if to_class is None:
            # Class name should by default start by upper case letter
            to_class = path[-1][0].isupper()

        # If caller doesn't want resolution, always return a Dummy path (except try_import module shortcut)
        if not to_class:
            if try_import:
                root = path[0]
                try:
                    return importlib.import_module(root)
                except ModuleNotFoundError:
                    return cls(path)
            return cls(path)

        # to_class=True: resolve as far as possible (module import + attribute walk).
        if not try_import:
            # Contract: if you didn't allow importing, we can't actually resolve anything.
            return cls(path)

        root = path[0]

        # Try to import the *longest* module prefix we can.
        # Example: ("a","b","C") -> try import "a.b" then getattr "C".
        mod = None
        imported_upto = 0
        for i in range(len(path), 0, -1):
            mod_name = ".".join(path[:i])
            try:
                mod = importlib.import_module(mod_name)
                imported_upto = i
                break
            except ModuleNotFoundError:
                continue

        # If nothing imports, bail to Dummy
        if mod is None:
            # One more attempt: import root only (helps when full prefix import fails weirdly)
            try:
                mod = importlib.import_module(root)
                imported_upto = 1
            except ModuleNotFoundError:
                return cls(path)

        # Walk remaining parts as attributes
        obj: Any = mod
        for name in path[imported_upto:]:
            try:
                obj = getattr(obj, name)
            except AttributeError:
                return cls(path)

        return obj

    def materialize(self, raise_error: bool = True) -> Any:
        """
        Try to import and resolve the dotted path.

        - If _path is ("pyarrow",) -> returns imported module
        - If _path is ("pyarrow","compute") -> tries:
            1) import "pyarrow.compute" as a module
            2) else import "pyarrow" then getattr "compute"
        - If anything fails:
            - raise_error=True  -> raise ModuleNotFoundError with nice message
            - raise_error=False -> return Dummy(self._path)
        """
        if not self._module_path:
            if raise_error:
                self._raise("import")
            return Dummy(("<unknown>",))

        root = self._module_path[0]
        target = self.dotted()

        # 1) Try importing the full dotted path directly (best case: it's a module)
        try:
            obj: Any = importlib.import_module(target)
        except ModuleNotFoundError:
            # 2) Fallback: import root, then resolve attributes
            try:
                obj = importlib.import_module(root)
            except ModuleNotFoundError:
                if raise_error:
                    self._raise("import")
                return Dummy(self._module_path)

            for attr in self._module_path[1:]:
                try:
                    obj = getattr(obj, attr)
                except AttributeError:
                    if raise_error:
                        self._raise("import", f"attribute '{attr}' not found")
                    return Dummy(self._module_path)

        return obj

    def dotted(self) -> str:
        return ".".join(self._module_path) if self._module_path else "<unknown>"

    def _raise(self, action: str, detail: str | None = None) -> Never:
        root = self._module_path[0] if self._module_path else "<unknown>"
        target = self.dotted()
        extra = f" ({detail})" if detail else ""
        raise ModuleNotFoundError(
            f"Missing optional dependency '{root}'. Tried to {action} on '{target}'{extra}."
        )

    # --- attribute access: reconstitute chain instead of failing ---
    def __getattr__(self, name: str) -> "Dummy":
        return Dummy((*self._module_path, name))

    # --- actual usage: fail loudly ---
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self._raise("call")

    def __bool__(self) -> bool:
        self._raise("coerce to bool")

    def __iter__(self):
        self._raise("iterate")

    def __len__(self) -> int:
        self._raise("get length")

    def __contains__(self, item: Any) -> bool:
        self._raise("check containment")

    def __getitem__(self, key: Any) -> Any:
        self._raise("index", repr(key))

    def __setitem__(self, key: Any, value: Any) -> None:
        self._raise("set item", repr(key))

    def __delitem__(self, key: Any) -> None:
        self._raise("delete item", repr(key))

    def __repr__(self) -> str:
        return f"<Dummy missing '{self.dotted()}'>"



from __future__ import annotations

from dataclasses import dataclass

from urllib3 import BaseHTTPResponse

from ..dynamic_buffer import DynamicBuffer
from ..request import PreparedRequest
from ..response import Response

__all__ = [
    "HTTPResponse"
]


@dataclass(slots=True)
class HTTPResponse(Response):

    @classmethod
    def from_urllib3(
        cls,
        request: "PreparedRequest",
        response: BaseHTTPResponse,
        stream: bool,
        received_at_timestamp: int
    ) -> "Response":
        buffer = DynamicBuffer()

        try:
            if stream:
                for batch in response.stream(amt=4 * 1024 * 1024):
                    buffer.write(batch)
            else:
                buffer.write(response.data)
        finally:
            # important when streaming / pooling
            response.release_conn()

        buffer.seek(0)

        return cls(
            request=request,
            status_code=response.status,
            headers=response.headers,
            buffer=buffer,
            received_at_timestamp=received_at_timestamp
        )



"""
Databricks SQL engine utilities and helpers.

This module provides a thin “do the right thing” layer over:
- Databricks SQL Statement Execution API (warehouse)
- Spark SQL / Delta Lake (when running inside a Spark-enabled context)

It includes helpers to:
- Build fully-qualified table names
- Execute SQL via Spark or Databricks SQL API
- Insert Arrow/Spark/tabular data into Delta tables (append/overwrite/merge)
- Generate DDL from Arrow schemas

Design goals:
- Prefer Spark execution when running inside Databricks with an active SparkSession.
- Otherwise, use the Warehouse Statement Execution API.
- Keep behavior consistent between Spark and SQL paths (especially MERGE semantics).
- Be safe by default: quote identifiers, escape SQL strings, validate column references.
"""

from __future__ import annotations

import dataclasses
import logging
import random
import string
import time
from threading import Thread
from typing import Optional, Union, Any, Dict, Literal, TYPE_CHECKING

import pyarrow as pa
from databricks.sdk.service.sql import Disposition
from yggdrasil.collections.expiring_dict import ExpiringDict
from yggdrasil.enums import SaveMode

from .exceptions import SqlStatementError
from .statement_result import StatementResult
from .types import column_info_to_arrow_field
from .warehouse import SQLWarehouse
from ..workspaces import WorkspaceService, DatabricksPath
from ...enums import FileFormat
from ...pyutils.waiting_config import WaitingConfigArg, WaitingConfig
from ...sql.engine import SQLEngine as BaseSQLEngine
from ...types import is_arrow_type_string_like, is_arrow_type_binary_like, arrow_field_to_schema
from ...types.cast.cast_options import CastOptions
from ...types.cast.registry import convert

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import pyspark
    import delta
    import pandas
    import polars

    from ...ai.sql_session import SQLAISession, SQLFlavor


__all__ = [
    "SQLEngine",
    "StatementResult",
    "CreateTablePlan",
]


_INVALID_COL_CHARS = set(" ,;{}()\n\t=")


def _escape_sql_string(s: str) -> str:
    """
    Escape a Python string for safe embedding in a single-quoted SQL literal.

    Args:
        s: Input string.

    Returns:
        SQL-escaped string where single quotes are doubled.
    """
    return s.replace("'", "''")


def _quote_ident(ident: str) -> str:
    """
    Quote a SQL identifier using backticks, escaping embedded backticks.

    Args:
        ident: Identifier to quote (catalog/schema/table/column/etc).

    Returns:
        Backtick-quoted identifier with embedded backticks doubled.
    """
    escaped = ident.replace("`", "``")
    return f"`{escaped}`"


def _needs_column_mapping(col_name: str) -> bool:
    """
    Heuristic: return True if a column name has characters that commonly break without
    Delta column mapping / strict quoting.

    Args:
        col_name: Column name.

    Returns:
        True if column name includes invalid/awkward characters.
    """
    return any(ch in _INVALID_COL_CHARS for ch in col_name)


@dataclasses.dataclass
class CreateTablePlan:
    """
    A plan produced by :meth:`SQLEngine.create_table`.

    Attributes:
        sql:
            The generated CREATE TABLE statement.
        properties:
            Final Delta/Databricks table properties emitted into TBLPROPERTIES.
        warnings:
            Non-fatal issues detected while building the statement (e.g., invalid column names
            while column mapping is disabled).
        arrow_field:
            Arrow field representing the table schema. Convention:
              - If struct: its children become table columns.
              - Otherwise: it represents a single-column table.
        result:
            The execution result when ``execute=True`` was used; otherwise None.
    """

    sql: str
    properties: dict[str, Any]
    warnings: list[str]
    arrow_field: pa.Field
    result: Optional[StatementResult] = None

    @property
    def arrow_schema(self) -> pa.Schema:
        """
        Return the table schema as an Arrow :class:`pyarrow.Schema`.

        Returns:
            Arrow schema derived from :attr:`arrow_field` using project conventions.
        """
        return arrow_field_to_schema(self.arrow_field, None)


@dataclasses.dataclass
class SQLEngine(BaseSQLEngine, WorkspaceService):
    """Execute SQL statements and manage tables via Databricks SQL / Spark."""
    catalog_name: Optional[str] = None
    schema_name: Optional[str] = None

    _warehouse: Optional[SQLWarehouse] = dataclasses.field(default=None, repr=False, hash=False, compare=False)
    _ai_session: Optional["SQLAISession"] = dataclasses.field(default=None, repr=False, hash=False, compare=False)
    _cached_queries: Optional[ExpiringDict[str, StatementResult]] = dataclasses.field(default=ExpiringDict, repr=False, hash=False, compare=False)

    # -------------------------------------------------------------------------------------
    # Naming helpers
    # -------------------------------------------------------------------------------------

    def table_full_name(
        self,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        safe_chars: bool = True,
    ) -> str:
        """
        Build a fully qualified table name (catalog.schema.table).

        Args:
            catalog_name:
                Optional catalog override (defaults to ``self.catalog_name``).
            schema_name:
                Optional schema override (defaults to ``self.schema_name``).
            table_name:
                Table name to qualify.
            safe_chars:
                If True, wraps each identifier in backticks.

        Returns:
            Fully qualified table name string.

        Raises:
            AssertionError: If any of catalog/schema/table is missing after defaults.
        """
        catalog_name = catalog_name or self.catalog_name
        schema_name = schema_name or self.schema_name

        assert catalog_name, "No catalog name given"
        assert schema_name, "No schema name given"
        assert table_name, "No table name given"

        if safe_chars:
            return f"{_quote_ident(catalog_name)}.{_quote_ident(schema_name)}.{_quote_ident(table_name)}"
        return f"{catalog_name}.{schema_name}.{table_name}"

    def _catalog_schema_table_names(self, full_name: str) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Parse a catalog.schema.table string into components.

        Supports partial names:
        - table
        - schema.table
        - catalog.schema.table

        Backticks are stripped from each part.

        Args:
            full_name:
                Fully qualified or partial table name string.

        Returns:
            Tuple (catalog_name, schema_name, table_name), where missing parts may be None
            and caller can fall back to engine defaults.
        """
        parts = [_.strip("`") for _ in full_name.split(".")]

        # Unreachable for .split("."), but harmless and keeps intent explicit.
        if len(parts) == 0:
            return self.catalog_name, self.schema_name, None
        if len(parts) == 1:
            return self.catalog_name, self.schema_name, parts[0]
        if len(parts) == 2:
            return self.catalog_name, parts[0], parts[1]

        catalog_name, schema_name, table_name = parts[-3], parts[-2], parts[-1]
        return catalog_name or self.catalog_name, schema_name or self.schema_name, table_name

    @staticmethod
    def _random_suffix(prefix: str = "") -> str:
        """
        Generate a unique suffix for temporary resources.

        Args:
            prefix: Optional prefix to put before the suffix.

        Returns:
            String of the form "{prefix}{timestamp_ms}_{random8}".
        """
        unique = "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(8))
        timestamp = int(time.time() * 1000)
        return f"{prefix}{timestamp}_{unique}"

    # -------------------------------------------------------------------------------------
    # Warehouse + AI session
    # -------------------------------------------------------------------------------------

    def warehouse(
        self,
        warehouse_id: Optional[str] = None,
        warehouse_name: Optional[str] = None,
    ) -> SQLWarehouse:
        """
        Resolve and cache a Databricks SQL warehouse.

        Behavior:
        - If no warehouse is cached, try to find one (including a starter warehouse).
        - If none found, create or update a warehouse according to workspace policy.
        - Always returns a resolved warehouse (or raises if resolution fails).

        Args:
            warehouse_id:
                Optional explicit warehouse ID to resolve.
            warehouse_name:
                Optional warehouse name to resolve.

        Returns:
            Resolved :class:`SQLWarehouse`.
        """
        if self._warehouse is None:
            wh = SQLWarehouse(
                workspace=self.workspace,
                warehouse_id=warehouse_id,
                warehouse_name=warehouse_name,
            )

            self._warehouse = wh.find_warehouse(
                warehouse_id=warehouse_id,
                warehouse_name=warehouse_name,
                raise_error=False,
                find_starter=True,
            )

            if self._warehouse is None:
                self._warehouse = wh.create_or_update()

        return self._warehouse.find_warehouse(
            warehouse_id=warehouse_id,
            warehouse_name=warehouse_name,
            raise_error=True,
        )

    def ai_session(
        self,
        model: str = "databricks-gemini-2-5-pro",
        flavor: Optional["SQLFlavor"] = None,
    ):
        """
        Create a SQL AI session (LLM-assisted SQL).

        Args:
            model:
                Serving endpoint/model name.
            flavor:
                SQL dialect/flavor. Defaults to Databricks.

        Returns:
            SQLAISession.
        """
        from ...ai.sql_session import SQLAISession, SQLFlavor

        if flavor is None:
            flavor = SQLFlavor.DATABRICKS

        return SQLAISession(
            model=model,
            api_key=self.workspace.current_token(),
            base_url="%s/serving-endpoints" % self.workspace.safe_host,
            flavor=flavor,
        )

    # -------------------------------------------------------------------------------------
    # Execution
    # -------------------------------------------------------------------------------------

    def execute(
        self,
        statement: str,
        *,
        row_limit: Optional[int] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        wait: Optional[WaitingConfigArg] = True,
        engine: Optional[Literal["spark", "api"]] = None,
        warehouse_id: Optional[str] = None,
        warehouse_name: Optional[str] = None,
        byte_limit: Optional[int] = None,
        cache_for: Optional[WaitingConfigArg] = None
    ) -> StatementResult:
        """
        Execute a SQL statement using either Spark SQL or the Databricks SQL Statement Execution API.

        Engine selection:
          - If `engine="spark"`, executes using the active SparkSession.
          - If `engine="api"`, executes using the SQL warehouse API.
          - If `engine` is None, defaults to Spark when running inside Databricks with a Spark environment;
            otherwise uses the warehouse API.

        Waiting behavior (API engine only):
          - `wait=True` blocks until the statement reaches a terminal state.
          - `wait=False` returns immediately with a handle; callers can wait later via `StatementResult.wait()`.

        Args:
            statement:
                SQL statement to execute.
            row_limit:
                Optional row limit applied (Spark: `df.limit()`, API: forwarded to API if supported).
            catalog_name:
                Optional catalog override for API engine context.
            schema_name:
                Optional schema override for API engine context.
            wait:
                Whether to wait for completion (API engine only).
            engine:
                Execution engine override: "spark" or "api".
            warehouse_id:
                Warehouse ID override (API engine).
            warehouse_name:
                Warehouse name override (API engine).
            byte_limit:
                Optional maximum bytes returned for results (API engine only).
            cache_for:
                Optional cache time

        Returns:
            StatementResult:
                Unified wrapper over either a Spark DataFrame result or a warehouse API statement result.

        Raises:
            ValueError:
                If `engine="spark"` is selected but there is no active SparkSession.
        """
        # --- Engine auto-detection ---
        if not engine:
            if self.workspace.is_in_databricks_environment():
                engine = "spark"

        statement = statement.strip()

        if cache_for is not None:
            cache_for = WaitingConfig.check_arg(cache_for)

            existing = self._cached_queries.get(statement)

            if existing is not None:
                return existing

        # --- Spark path ---
        if engine == "spark":
            from ...spark.lib import pyspark_sql

            spark_session = pyspark_sql.SparkSession.getActiveSession()
            if spark_session is None:
                raise ValueError("No spark session found to run sql query")

            logger.debug("SPARK SQL executing query:\n%s", statement)

            df: pyspark_sql.DataFrame = spark_session.sql(statement)
            if row_limit:
                df = df.limit(row_limit)

            result = StatementResult(
                workspace_client=self.workspace.sdk(),
                warehouse_id="SparkSQL",
                statement_id="SparkSQL",
                disposition=Disposition.EXTERNAL_LINKS,
            )

            result._spark_df = df
        else:
            # --- Warehouse API path ---
            wh = self.warehouse(
                warehouse_id=warehouse_id,
                warehouse_name=warehouse_name,
            )

            result = wh.execute(
                statement=statement,
                warehouse_id=warehouse_id,
                warehouse_name=warehouse_name,
                byte_limit=byte_limit,
                catalog_name=catalog_name,
                schema_name=schema_name,
                wait=wait,
            )

        if cache_for is not None:
            self._cached_queries.set(
                key=statement, value=result,
                ttl=cache_for.timeout_total_seconds
            )

        return result

    # -------------------------------------------------------------------------------------
    # Delta table handle (Spark required)
    # -------------------------------------------------------------------------------------

    def spark_table(
        self,
        full_name: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
    ) -> "delta.tables.DeltaTable":
        """
        Return a DeltaTable handle for a given table name (Spark context required).

        Args:
            full_name:
                Fully qualified table name (catalog.schema.table). If provided, takes precedence.
            catalog_name:
                Catalog name used when `full_name` is not provided.
            schema_name:
                Schema name used when `full_name` is not provided.
            table_name:
                Table name used when `full_name` is not provided.

        Returns:
            delta.tables.DeltaTable.

        Raises:
            ImportError:
                If delta-spark cannot be imported and runtime install fails.
        """
        from ...spark.lib import pyspark_sql

        try:
            from delta.tables import DeltaTable
        except ImportError:
            from ...pyutils.pyenv import PyEnv

            m = PyEnv.runtime_import_module(module_name="delta.tables", pip_name="delta-spark", install=True)
            DeltaTable = m.DeltaTable

        if not full_name:
            full_name = self.table_full_name(
                catalog_name=catalog_name,
                schema_name=schema_name,
                table_name=table_name,
            )

        return DeltaTable.forName(
            sparkSession=pyspark_sql.SparkSession.getActiveSession(),
            tableOrViewName=full_name,
        )

    # -------------------------------------------------------------------------------------
    # Insert routing
    # -------------------------------------------------------------------------------------

    def insert_into(
        self,
        data: Union[
            pa.Table,
            pa.RecordBatch,
            pa.RecordBatchReader,
            dict,
            list,
            str,
            "pandas.DataFrame",
            "polars.DataFrame",
            "pyspark.sql.DataFrame",
        ],
        *,
        mode: SaveMode | str | None = None,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        cast_options: Optional[CastOptions] = None,
        overwrite_schema: bool | None = None,
        match_by: Optional[list[str]] = None,
        wait: Optional[WaitingConfigArg] = True,
        ## Databricks specific
        zorder_by: Optional[list[str]] = None,
        optimize_after_merge: bool = False,
        vacuum_hours: int | None = None,
        spark_session: Optional["pyspark.sql.SparkSession"] = None,
        spark_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Insert data into a Delta table.

        Routing strategy:
          - If a SparkSession is available (or `data` is a Spark DataFrame), uses :meth:`spark_insert_into`.
          - Otherwise uses :meth:`arrow_insert_into` which stages Parquet to a temp volume and executes SQL.

        Args:
            data:
                Input data. Supported (path-dependent) inputs include:
                  - Arrow: `pa.Table`, `pa.RecordBatch`, `pa.RecordBatchReader`
                  - Spark: `pyspark.sql.DataFrame`
                  - Tabular Python: dict/list, pandas.DataFrame, polars.DataFrame (convertible)
                  - str: path or special locator (only if your conversion utilities support it)
            mode:
                Save mode/semantics (auto/append/overwrite). Parsed via `SaveMode.from_any`.
            location:
                Fully qualified table name override (catalog.schema.table).
            catalog_name:
                Optional catalog override (used when `location` not provided or partial).
            schema_name:
                Optional schema override (used when `location` not provided or partial).
            table_name:
                Optional table name (used when `location` not provided).
            cast_options:
                Casting rules for aligning input columns to the destination schema.
            overwrite_schema:
                Spark-only: if True, sets write option `overwriteSchema=true`.
            match_by:
                Merge key columns. Enables MERGE semantics (upsert/insert-only/overwrite-by-key).
            wait:
                Waiting completion configuration
            zorder_by:
                Columns for Z-ORDER optimization after write (path-dependent support).
            optimize_after_merge:
                If True, run OPTIMIZE after merge (and ZORDER where supported).
            vacuum_hours:
                If set, run VACUUM with the given retention (hours).
            spark_session:
                Optional SparkSession override. If not provided, uses active session when available.
            spark_options:
                Optional Spark DataFrameWriter options.

        Returns:
            None.
        """
        if spark_session is None:
            try:
                import pyspark.sql as psql

                if isinstance(data, psql.DataFrame):
                    spark_session = data.sparkSession
                else:
                    spark_session = psql.SparkSession.getActiveSession()
            except ImportError:
                pass

        if spark_session is not None:
            return self.spark_insert_into(
                data=data,
                location=location,
                catalog_name=catalog_name,
                schema_name=schema_name,
                table_name=table_name,
                mode=mode,
                cast_options=cast_options,
                overwrite_schema=overwrite_schema,
                match_by=match_by,
                zorder_by=zorder_by,
                optimize_after_merge=optimize_after_merge,
                vacuum_hours=vacuum_hours,
                spark_options=spark_options,
            )
        else:
            return self.arrow_insert_into(
                data=data,
                location=location,
                catalog_name=catalog_name,
                schema_name=schema_name,
                table_name=table_name,
                mode=mode,
                cast_options=cast_options,
                overwrite_schema=overwrite_schema,
                match_by=match_by,
                wait=wait,
                zorder_by=zorder_by,
                optimize_after_merge=optimize_after_merge,
                vacuum_hours=vacuum_hours,
            )

    # -------------------------------------------------------------------------------------
    # Warehouse SQL insert path
    # -------------------------------------------------------------------------------------

    def arrow_insert_into(
        self,
        data: Union[
            pa.Table,
            pa.RecordBatch,
            pa.RecordBatchReader,
            dict,
            list,
            "pandas.DataFrame",
            "polars.DataFrame",
        ],
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        mode: SaveMode | str | None = None,
        cast_options: Optional[CastOptions] = None,
        overwrite_schema: bool | None = None,
        match_by: Optional[list[str]] = None,
        wait: Optional[WaitingConfigArg] = True,
        zorder_by: Optional[list[str]] = None,
        optimize_after_merge: bool = False,
        vacuum_hours: int | None = None,
        existing_schema: pa.Schema | None = None,
        temp_volume_path: Optional[Union[str, DatabricksPath]] = None,
    ) -> None:
        """
        Insert data using the warehouse SQL path by staging Parquet files to a temporary volume, then executing
        SQL INSERT/MERGE into the target Delta table.

        Notes:
          - If the table does not exist, it is created from the input schema (best-effort) via :meth:`create_table`.
          - If `match_by` is provided, uses MERGE semantics:
              * mode=overwrite: delete matching keys, then insert batch (overwrite-by-key)
              * mode=append: insert-only (no updates)
              * mode=auto/other: full upsert (update matched + insert new)
          - If `match_by` is not provided:
              * mode=overwrite: INSERT OVERWRITE
              * otherwise: INSERT INTO
          - Rows with NULL in any match key are filtered out (mirrors Spark path).

        Args:
            data:
                Input data (Arrow or convertible). Non-Arrow inputs may be converted by project utilities.
            location:
                Fully qualified table name override (catalog.schema.table).
            catalog_name:
                Optional catalog override (used when `location` not provided/partial).
            schema_name:
                Optional schema override (used when `location` not provided/partial).
            table_name:
                Optional table name (used when `location` not provided).
            mode:
                Save mode/semantics (auto/append/overwrite) parsed via `SaveMode.from_any`.
            cast_options:
                Casting rules for aligning staged Parquet columns to destination schema.
            overwrite_schema:
                Reserved for parity with Spark path (unused in SQL staging path).
            match_by:
                Merge key columns enabling MERGE behavior.
            wait:
                Waiting completion configuration
            zorder_by:
                If set, runs `OPTIMIZE ... ZORDER BY (...)` after inserting.
            optimize_after_merge:
                If True and `match_by` is set, runs `OPTIMIZE <table>` after merge (in addition to ZORDER).
            vacuum_hours:
                If set, runs `VACUUM <table> RETAIN <hours> HOURS`.
            existing_schema:
                Optional destination schema (Arrow) to avoid an extra schema fetch.
            temp_volume_path:
                Optional staging path override. If not provided, a temp volume path is allocated.

        Returns:
            None.
        """
        location, catalog_name, schema_name, table_name = self._check_location_params(
            location=location,
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=True,
        )

        with self.connect() as connected:
            if existing_schema is None:
                try:
                    existing_schema = connected.get_table_schema(
                        location=location,
                        catalog_name=catalog_name,
                        schema_name=schema_name,
                        table_name=table_name,
                        to_arrow_schema=True,
                    )
                except ValueError as exc:
                    if isinstance(data, (list, dict)):
                        from ...polars.cast import any_to_polars_dataframe

                        data = any_to_polars_dataframe(data, cast_options)

                    logger.warning("%s, creating it from input schema %s", exc, repr(data))

                    plan = connected.create_table(
                        field=data,  # convertible to pa.Field via convert() in create_table
                        catalog_name=catalog_name,
                        schema_name=schema_name,
                        table_name=table_name,
                        if_not_exists=True,
                        execute=True,
                    )

                    try:
                        connected.arrow_insert_into(
                            data=data,
                            location=location,
                            catalog_name=catalog_name,
                            schema_name=schema_name,
                            table_name=table_name,
                            mode="overwrite",
                            cast_options=cast_options,
                            overwrite_schema=overwrite_schema,
                            match_by=match_by,
                            wait=wait,
                            zorder_by=zorder_by,
                            optimize_after_merge=optimize_after_merge,
                            vacuum_hours=vacuum_hours,
                            existing_schema=plan.arrow_schema,
                        )
                        return
                    except Exception:
                        logger.exception(
                            "Arrow insert failed after auto-creating %s; attempting cleanup (DROP TABLE)", location
                        )
                        try:
                            connected.drop_table(location=location, wait=True)
                        except Exception as e:
                            logger.exception("Failed to drop table %s after auto creation error: %s", location, e)
                        raise

            cast_options = CastOptions.check_arg(options=cast_options, target_field=existing_schema)

            logger.debug("Inserting %s into %s", data, location)

            temp_volume_path = (
                self.workspace.tmp_path(
                    catalog_name=catalog_name,
                    schema_name=schema_name,
                    volume_name="tmp",
                    extension="parquet",
                    max_lifetime=3600,
                )
                if temp_volume_path is None
                else DatabricksPath.parse(obj=temp_volume_path, workspace=connected.workspace)
            )

            temp_volume_path.write_table(
                data,
                file_format=FileFormat.PARQUET,
                cast_options=cast_options,
            )

            mode = SaveMode.from_any(mode, default=SaveMode.AUTO)

            columns = list(existing_schema.names)
            cols_quoted = ", ".join([_quote_ident(c) for c in columns])

            statements: list[str] = []

            if match_by:
                not_null_pred = " AND ".join([f"{_quote_ident(k)} IS NOT NULL" for k in match_by])

                source_sql = f"""SELECT {cols_quoted}
FROM parquet.{_quote_ident(str(temp_volume_path))}
WHERE {not_null_pred}""".strip()

                on_condition = " AND ".join([f"T.{_quote_ident(k)} = S.{_quote_ident(k)}" for k in match_by])

                if mode == SaveMode.OVERWRITE:
                    key_cols = ", ".join([_quote_ident(k) for k in match_by])

                    delete_sql = f"""DELETE FROM {location} AS T
USING (
  SELECT DISTINCT {key_cols}
  FROM parquet.{_quote_ident(str(temp_volume_path))}
  WHERE {not_null_pred}
) AS S
ON {on_condition}""".strip()

                    insert_sql = f"""INSERT INTO {location} ({cols_quoted})
{source_sql}""".strip()

                    # Need consecutive queries
                    wait = True
                    statements.extend([delete_sql, insert_sql])

                elif mode == SaveMode.APPEND:
                    insert_clause = (
                        f"WHEN NOT MATCHED THEN INSERT ({cols_quoted}) "
                        f"VALUES ({', '.join([f'S.{_quote_ident(c)}' for c in columns])})"
                    )

                    merge_sql = f"""MERGE INTO {location} AS T
USING (
  {source_sql}
) AS S
ON {on_condition}
{insert_clause}""".strip()

                    statements.append(merge_sql)

                else:
                    update_cols = [c for c in columns if c not in match_by]
                    update_clause = ""
                    if update_cols:
                        update_set = ", ".join([f"T.{_quote_ident(c)} = S.{_quote_ident(c)}" for c in update_cols])
                        update_clause = f"WHEN MATCHED THEN UPDATE SET {update_set}"

                    insert_clause = (
                        f"WHEN NOT MATCHED THEN INSERT ({cols_quoted}) "
                        f"VALUES ({', '.join([f'S.{_quote_ident(c)}' for c in columns])})"
                    )

                    merge_sql = f"""MERGE INTO {location} AS T
USING (
  {source_sql}
) AS S
ON {on_condition}
{update_clause}
{insert_clause}""".strip()

                    statements.append(merge_sql)

            else:
                if mode == SaveMode.OVERWRITE:
                    insert_sql = f"""INSERT OVERWRITE {location}
SELECT {cols_quoted}
FROM parquet.{_quote_ident(str(temp_volume_path))}"""
                else:
                    insert_sql = f"""INSERT INTO {location} ({cols_quoted})
SELECT {cols_quoted}
FROM parquet.{_quote_ident(str(temp_volume_path))}"""
                statements.append(insert_sql)

            try:
                for stmt in statements:
                    connected.execute(stmt, wait=wait)
            finally:
                try:
                    Thread(target=temp_volume_path.remove, kwargs={"recursive": True}).start()
                except Exception:
                    logger.exception("Failed cleaning temp volume: %s", temp_volume_path)

            logger.info("Arrow inserted into %s", location)

            if zorder_by:
                zorder_cols = ", ".join([_quote_ident(c) for c in zorder_by])
                connected.execute(f"OPTIMIZE {location} ZORDER BY ({zorder_cols})")

            if optimize_after_merge and match_by:
                connected.execute(f"OPTIMIZE {location}")

            if vacuum_hours is not None:
                connected.execute(f"VACUUM {location} RETAIN {int(vacuum_hours)} HOURS")

        return None

    # -------------------------------------------------------------------------------------
    # Spark insert path
    # -------------------------------------------------------------------------------------

    def spark_insert_into(
        self,
        data: Any,
        *,
        mode: SaveMode | str | None = None,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        cast_options: Optional[CastOptions] = None,
        overwrite_schema: bool | None = None,
        match_by: Optional[list[str]] = None,
        zorder_by: Optional[list[str]] = None,
        optimize_after_merge: bool = False,
        vacuum_hours: int | None = None,
        spark_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Insert data into a Delta table using Spark.

        Accepts a Spark DataFrame or any input convertible to a Spark DataFrame via
        your project's `any_to_spark_dataframe()` utility.

        Behavior:
          - If destination table does not exist: creates it via `saveAsTable(mode="overwrite")`.
          - If `match_by` is provided: uses Delta MERGE semantics.
              * mode=overwrite: overwrite-by-key (delete matching keys, then append)
              * mode=append: insert-only (whenNotMatchedInsertAll)
              * mode=auto/other: upsert (whenMatchedUpdate + whenNotMatchedInsertAll)
          - If no `match_by`: uses DataFrameWriter `saveAsTable()` with append/overwrite.

        Notes:
          - Rows with NULL in any match key are filtered out (mirrors SQL path).
          - `optimize_after_merge` will only run ZORDER when `zorder_by` is provided.
          - If `overwrite_schema` is True, sets Spark option `overwriteSchema=true`.

        Args:
            data:
                Input data convertible to Spark DataFrame.
            mode:
                Save mode/semantics (auto/append/overwrite) parsed via `SaveMode.from_any`.
            location:
                Fully qualified table name override (catalog.schema.table).
            catalog_name:
                Optional catalog override (used when `location` not provided/partial).
            schema_name:
                Optional schema override (used when `location` not provided/partial).
            table_name:
                Optional table name (used when `location` not provided).
            cast_options:
                Casting rules for aligning input columns to destination schema.
            overwrite_schema:
                If True, sets writer option `overwriteSchema=true` (Spark path).
            match_by:
                Merge key columns enabling MERGE behavior.
            zorder_by:
                Z-ORDER columns for optimize after merge (requires `optimize_after_merge=True`).
            optimize_after_merge:
                If True and `zorder_by` is provided, runs Delta optimize ZORDER after merge/write.
            vacuum_hours:
                If set, runs Delta vacuum with given retention hours.
            spark_options:
                Spark DataFrameWriter options dict (e.g., {"mergeSchema": "true"}).

        Returns:
            None.
        """
        from ...spark.cast import any_to_spark_dataframe
        from pyspark.sql import DataFrame
        import pyspark.sql.functions as F

        location, catalog_name, schema_name, table_name = self._check_location_params(
            location=location,
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=True,
        )

        logger.info(
            "Spark insert into %s (mode=%s, match_by=%s, overwrite_schema=%s)",
            location,
            mode,
            match_by,
            overwrite_schema,
        )

        spark_options = spark_options if spark_options else {}
        if overwrite_schema:
            spark_options["overwriteSchema"] = "true"

        try:
            existing_schema = self.get_table_schema(
                catalog_name=catalog_name,
                schema_name=schema_name,
                table_name=table_name,
                to_arrow_schema=False,
            )
        except ValueError:
            logger.warning("Destination table missing; creating table %s via overwrite write", location)
            data_df = convert(data, DataFrame)
            data_df.write.mode("overwrite").options(**spark_options).saveAsTable(location)
            return

        cast_options = CastOptions.check_arg(options=cast_options, target_field=existing_schema)
        data_df = any_to_spark_dataframe(data, cast_options)

        if match_by:
            notnull = None
            for k in match_by:
                if k not in data_df.columns:
                    raise ValueError(f"Missing match key '{k}' in DataFrame columns: {data_df.columns}")
                notnull = data_df[k].isNotNull() if notnull is None else notnull & data_df[k].isNotNull()

            data_df = data_df.filter(notnull)
            logger.debug("Filtered null keys for match_by=%s", match_by)

        target = self.spark_table(full_name=location)
        mode = SaveMode.from_any(mode, default=SaveMode.AUTO)

        if match_by:
            cond = " AND ".join([f"t.{_quote_ident(k)} <=> s.{_quote_ident(k)}" for k in match_by])

            if mode == SaveMode.OVERWRITE:
                data_df = data_df.cache()
                distinct_keys = data_df.select([f"{_quote_ident(k)}" for k in match_by]).distinct()

                (
                    target.alias("t")
                    .merge(distinct_keys.alias("s"), cond)
                    .whenMatchedDelete()
                    .execute()
                )

                (
                    data_df.write.format("delta")
                    .mode("append")
                    .options(**spark_options)
                    .saveAsTable(location)
                )

            elif mode == SaveMode.APPEND:
                (
                    target.alias("t")
                    .merge(data_df.alias("s"), cond)
                    .whenNotMatchedInsertAll()
                    .execute()
                )

            else:
                update_cols = [c for c in data_df.columns if c not in match_by]
                set_expr = {c: F.expr(f"s.{_quote_ident(c)}") for c in update_cols}

                (
                    target.alias("t")
                    .merge(data_df.alias("s"), cond)
                    .whenMatchedUpdate(set=set_expr)
                    .whenNotMatchedInsertAll()
                    .execute()
                )

        else:
            spark_mode = "overwrite" if mode == SaveMode.OVERWRITE else "append"
            logger.info("Spark write saveAsTable mode=%s", spark_mode)
            data_df.write.mode(spark_mode).options(**spark_options).saveAsTable(location)

        if optimize_after_merge and zorder_by:
            logger.info("Delta optimize + zorder (%s)", zorder_by)
            target.optimize().executeZOrderBy(*zorder_by)

        if vacuum_hours is not None:
            logger.info("Delta vacuum retain=%s hours", vacuum_hours)
            target.vacuum(vacuum_hours)

        return None

    # -------------------------------------------------------------------------------------
    # Schema + table management
    # -------------------------------------------------------------------------------------

    def get_table_schema(
        self,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        to_arrow_schema: bool = True,
    ) -> Union[pa.Field, pa.Schema]:
        """
        Fetch a table schema from Unity Catalog and convert it to Arrow types.

        Args:
            location:
                Optional fully qualified table name (catalog.schema.table).
            catalog_name:
                Optional catalog override (used when `location` not provided/partial).
            schema_name:
                Optional schema override (used when `location` not provided/partial).
            table_name:
                Optional table name (used when `location` not provided).
            to_arrow_schema:
                If True returns a `pa.Schema`.
                If False returns a `pa.Field` with STRUCT type representing the table schema.

        Returns:
            `pa.Schema` or `pa.Field` depending on `to_arrow_schema`.

        Raises:
            ValueError:
                If the table cannot be retrieved from Unity Catalog.
        """
        location, catalog_name, schema_name, table_name = self._check_location_params(
            location=location,
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=False,
        )

        client = self.workspace.sdk().tables

        try:
            table = client.get(location)
        except Exception as e:
            raise ValueError(f"Table {location} not found, {type(e)} {e}")

        fields = [column_info_to_arrow_field(_) for _ in table.columns]

        metadata = {
            b"engine": b"databricks",
            b"full_name": location.encode("utf-8"),
            b"catalog_name": (catalog_name or "").encode("utf-8"),
            b"schema_name": (schema_name or "").encode("utf-8"),
            b"table_name": (table_name or "").encode("utf-8"),
        }

        if to_arrow_schema:
            return pa.schema(fields, metadata=metadata)

        return pa.field(location, pa.struct(fields), metadata=metadata)

    def drop_table(
        self,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        wait: Optional[WaitingConfigArg] = True,
    ) -> None:
        """
        Drop a table if it exists.

        Args:
            location:
                Optional fully qualified table name (catalog.schema.table).
            catalog_name:
                Optional catalog override (used when `location` not provided/partial).
            schema_name:
                Optional schema override (used when `location` not provided/partial).
            table_name:
                Optional table name (used when `location` not provided).
            wait:
                Whether to block until the DROP finishes (API engine only).

        Returns:
            None.
        """
        location, _, _, _ = self._check_location_params(
            location=location,
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=True,
        )

        logger.debug("Dropping table if exists: %s", location)
        self.execute(f"DROP TABLE IF EXISTS {location}", wait=wait)
        logger.info("Dropped table if exists: %s", location)

    def create_table(
        self,
        field: Union[pa.Field, pa.Schema, Any],
        full_name: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        storage_location: Optional[str] = None,
        partition_by: Optional[list[str]] = None,
        cluster_by: Optional[bool | list[str]] = True,
        comment: Optional[str] = None,
        properties: Optional[dict[str, Any]] = None,
        if_not_exists: bool = True,
        or_replace: bool = False,
        using: str = "DELTA",
        optimize_write: bool = True,
        auto_compact: bool = True,
        enable_cdf: Optional[bool] = None,
        enable_deletion_vectors: Optional[bool] = None,
        target_file_size: Optional[int] = None,
        column_mapping_mode: Optional[str] = None,
        execute: bool = True,
        wait_result: bool = True,
    ) -> CreateTablePlan:
        """
        Generate (and optionally execute) a Databricks/Delta CREATE TABLE statement from an Apache Arrow schema/field,
        with safety checks and performance-oriented defaults.

        Safety/perf behaviors:
          - Quotes identifiers to avoid keyword/name edge cases.
          - Validates `partition_by` / `cluster_by` columns exist in schema.
          - Supports managed or external tables via `storage_location`.
          - Optionally enables Delta Column Mapping (name/id) with required protocol props.
          - Adds workspace default tags into table properties (`tags.<k>`).

        Args:
            field:
                Arrow schema/field describing the table, or any object convertible to `pa.Field` via `convert()`.
                - If `pa.Schema`: all schema fields become columns.
                - If struct `pa.Field`: its children become columns.
                - If non-struct `pa.Field`: single-column table.
            full_name:
                Fully qualified table name "catalog.schema.table". If provided, takes precedence.
            catalog_name:
                Catalog used when `full_name` not provided (or as override for partial `full_name`).
            schema_name:
                Schema used when `full_name` not provided (or as override for partial `full_name`).
            table_name:
                Table used when `full_name` not provided (or as override for partial `full_name`).
            storage_location:
                External storage location path. If set, emits `LOCATION '<path>'` (SQL-escaped).
            partition_by:
                Partition column names. Must exist in schema.
                Note: if set, clustering is not emitted (partition wins).
            cluster_by:
                Controls clustering / liquid clustering.
                - True: emits `CLUSTER BY AUTO`
                - False: emits nothing
                - list[str]: emits `CLUSTER BY (<cols...>)` (must exist in schema)
                Note: only applied when `partition_by` is not set.
            comment:
                Table comment. If None and Arrow metadata contains b"comment", that is used.
            properties:
                Additional/override Delta table properties (caller wins last).
            if_not_exists:
                If True, generates `CREATE TABLE IF NOT EXISTS ...`. Mutually exclusive with `or_replace`.
            or_replace:
                If True, generates `CREATE OR REPLACE TABLE ...`. Mutually exclusive with `if_not_exists`.
            using:
                Storage format keyword (default "DELTA").
            optimize_write:
                Sets `delta.autoOptimize.optimizeWrite`.
            auto_compact:
                Sets `delta.autoOptimize.autoCompact`.
            enable_cdf:
                If set, sets `delta.enableChangeDataFeed`.
            enable_deletion_vectors:
                If set, sets `delta.enableDeletionVectors`.
            target_file_size:
                If set, sets `delta.targetFileSize` (bytes).
            column_mapping_mode:
                Delta column mapping mode:
                - None: auto-detect (enable "name" iff invalid column names exist, else "none")
                - "none": do not enable column mapping
                - "name": enable name-based column mapping
                - "id": enable id-based column mapping
                When enabled (name/id), also sets:
                - `delta.minReaderVersion = 2`
                - `delta.minWriterVersion = 5`
            execute:
                If True, executes the generated SQL via :meth:`execute`.
            wait_result:
                Passed to :meth:`execute` when `execute=True`.

        Returns:
            CreateTablePlan:
                Always returns a plan. If `execute=True`, plan.result is populated.

        Raises:
            ValueError:
                On invalid naming parameters, conflicting flags, invalid column mapping mode, or missing columns.
            SqlStatementError:
                If execution fails and cannot be recovered (e.g., schema creation retry fails).
        """
        if not isinstance(field, pa.Field):
            field = convert(field, pa.Field)

        schema_metadata = field.metadata or {}

        if pa.types.is_struct(field.type):
            arrow_fields = list(field.type)
        else:
            arrow_fields = [field]

        full_name, catalog_name, schema_name, table_name = self._check_location_params(
            location=full_name,
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=True,
        )

        if comment is None and schema_metadata:
            c = schema_metadata.get(b"comment")
            if isinstance(c, bytes):
                comment = c.decode("utf-8")

        any_invalid = any(_needs_column_mapping(f.name) for f in arrow_fields)
        warnings: list[str] = []

        if column_mapping_mode is None:
            column_mapping_mode = "name" if any_invalid else "none"

        if column_mapping_mode not in ("none", "name", "id"):
            raise ValueError("column_mapping_mode must be one of: None, 'none', 'name', 'id'.")

        col_names = {f.name for f in arrow_fields}

        if partition_by:
            missing = [c for c in partition_by if c not in col_names]
            if missing:
                raise ValueError(f"partition_by contains unknown columns: {missing}")

        if isinstance(cluster_by, list):
            missing = [c for c in cluster_by if c not in col_names]
            if missing:
                raise ValueError(f"cluster_by contains unknown columns: {missing}")

        column_definitions = [self._field_to_ddl(child) for child in arrow_fields]

        if or_replace and if_not_exists:
            raise ValueError("Use either or_replace or if_not_exists, not both.")

        create_kw = "CREATE OR REPLACE TABLE" if or_replace else "CREATE TABLE"
        if if_not_exists and not or_replace:
            create_kw = "CREATE TABLE IF NOT EXISTS"

        sql_parts: list[str] = [
            f"{create_kw} {full_name} (",
            "  " + ",\n  ".join(column_definitions),
            ")",
            f"USING {using}",
        ]

        if partition_by:
            sql_parts.append("PARTITIONED BY (" + ", ".join(_quote_ident(c) for c in partition_by) + ")")
        elif cluster_by:
            if isinstance(cluster_by, bool):
                if cluster_by:
                    sql_parts.append("CLUSTER BY AUTO")
            else:
                sql_parts.append("CLUSTER BY (" + ", ".join(_quote_ident(c) for c in cluster_by) + ")")

        if comment:
            sql_parts.append(f"COMMENT '{_escape_sql_string(comment)}'")

        if storage_location:
            sql_parts.append(f"LOCATION '{_escape_sql_string(storage_location)}'")

        props: dict[str, Any] = {
            "delta.autoOptimize.optimizeWrite": bool(optimize_write),
            "delta.autoOptimize.autoCompact": bool(auto_compact),
        }

        if enable_cdf is not None:
            props["delta.enableChangeDataFeed"] = bool(enable_cdf)

        if enable_deletion_vectors is not None:
            props["delta.enableDeletionVectors"] = bool(enable_deletion_vectors)

        if target_file_size is not None:
            props["delta.targetFileSize"] = int(target_file_size)

        if column_mapping_mode != "none":
            props["delta.columnMapping.mode"] = column_mapping_mode
            props["delta.minReaderVersion"] = 2
            props["delta.minWriterVersion"] = 5

        if properties:
            props.update(properties)

        if any_invalid and column_mapping_mode == "none":
            warnings.append(
                "Schema has invalid column names but column_mapping_mode='none'. "
                "This will fail unless you rename/escape columns or enable column mapping."
            )

        default_tags = self.workspace.default_tags()
        for k, v in default_tags.items():
            props[f"tags.{k}"] = v

        if props:

            def fmt(_key: str, _value: Any) -> str:
                if isinstance(_value, str):
                    return f"'{_key}' = '{_escape_sql_string(_value)}'"
                if isinstance(_value, bool):
                    return f"'{_key}' = '{'true' if _value else 'false'}'"
                return f"'{_key}' = {_value}"

            sql_parts.append("TBLPROPERTIES (" + ", ".join(fmt(k, v) for k, v in props.items()) + ")")

        statement = "\n".join(sql_parts)
        plan = CreateTablePlan(sql=statement, properties=props, warnings=warnings, arrow_field=field)

        if not execute:
            return plan

        try:
            res = self.execute(statement, wait=wait_result)
        except SqlStatementError as e:
            if "SCHEMA_NOT_FOUND" in e.message:
                self.execute(f"CREATE SCHEMA IF NOT EXISTS {_quote_ident(schema_name)}", wait=True)
                res = self.execute(statement, wait=wait_result)
            else:
                raise

        plan.result = res
        return plan

    def _check_location_params(
        self,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        safe_chars: bool = True,
    ) -> tuple[str, Optional[str], Optional[str], Optional[str]]:
        """
        Resolve (location OR catalog/schema/table) into a fully-qualified table identifier.

        Args:
            location:
                Fully or partially-qualified table name string. Accepts:
                - table
                - schema.table
                - catalog.schema.table
            catalog_name:
                Optional catalog override, or default when `location` doesn't provide it.
            schema_name:
                Optional schema override, or default when `location` doesn't provide it.
            table_name:
                Optional table override, or default when `location` doesn't provide it.
            safe_chars:
                If True, quotes identifiers using backticks.

        Returns:
            Tuple (location_fqn, catalog_name, schema_name, table_name) where:
            - location_fqn is the resolved full name (quoted if safe_chars=True)
            - the name components are resolved with engine defaults as needed
        """
        if location:
            c, s, t = self._catalog_schema_table_names(location)
            catalog_name, schema_name, table_name = catalog_name or c, schema_name or s, table_name or t

        location = self.table_full_name(
            catalog_name=catalog_name,
            schema_name=schema_name,
            table_name=table_name,
            safe_chars=safe_chars,
        )

        return location, catalog_name or self.catalog_name, schema_name or self.schema_name, table_name

    @staticmethod
    def _field_to_ddl(
        field: pa.Field,
        put_name: bool = True,
        put_not_null: bool = True,
        put_comment: bool = True,
    ) -> str:
        """
        Convert an Arrow field to a Databricks SQL column DDL fragment.

        Supports primitives plus nested types:
        - STRUCT<...>
        - MAP<k,v>
        - ARRAY<elem>

        Args:
            field:
                Arrow field to convert.
            put_name:
                If True, include the column name.
            put_not_null:
                If True, emit NOT NULL when `field.nullable` is False.
            put_comment:
                If True, emit COMMENT when field metadata contains b"comment".

        Returns:
            SQL fragment for a column definition.

        Raises:
            TypeError:
                If a nested Arrow type cannot be represented.
            ValueError:
                If a primitive Arrow type cannot be mapped to a SQL type.
        """
        name = field.name
        nullable_str = " NOT NULL" if put_not_null and not field.nullable else ""
        name_str = f"{_quote_ident(name)} " if put_name else ""

        comment_str = ""
        if put_comment and field.metadata and b"comment" in field.metadata:
            comment = field.metadata[b"comment"].decode("utf-8")
            comment_str = f" COMMENT '{_escape_sql_string(comment)}'"

        if not pa.types.is_nested(field.type):
            sql_type = SQLEngine._arrow_to_sql_type(field.type)
            return f"{name_str}{sql_type}{nullable_str}{comment_str}"

        if pa.types.is_struct(field.type):
            struct_body = ", ".join([SQLEngine._field_to_ddl(child) for child in field.type])
            return f"{name_str}STRUCT<{struct_body}>{nullable_str}{comment_str}"

        if pa.types.is_map(field.type):
            map_type: pa.MapType = field.type
            key_type = SQLEngine._field_to_ddl(
                map_type.key_field,
                put_name=False,
                put_comment=False,
                put_not_null=False,
            )
            val_type = SQLEngine._field_to_ddl(
                map_type.item_field,
                put_name=False,
                put_comment=False,
                put_not_null=False,
            )
            return f"{name_str}MAP<{key_type}, {val_type}>{nullable_str}{comment_str}"

        if pa.types.is_list(field.type) or pa.types.is_large_list(field.type):
            list_type: pa.ListType = field.type
            elem_type = SQLEngine._field_to_ddl(
                list_type.value_field,
                put_name=False,
                put_comment=False,
                put_not_null=False,
            )
            return f"{name_str}ARRAY<{elem_type}>{nullable_str}{comment_str}"

        raise TypeError(f"Cannot make ddl field from {field}")

    @staticmethod
    def _arrow_to_sql_type(arrow_type: Union[pa.DataType, pa.Decimal128Type]) -> str:
        """
        Convert an Arrow data type to a Databricks SQL type string.

        Args:
            arrow_type:
                Arrow type instance to convert.

        Returns:
            Databricks SQL type string.

        Raises:
            ValueError:
                If the Arrow type is unsupported.
        """
        if pa.types.is_boolean(arrow_type):
            return "BOOLEAN"
        if pa.types.is_int8(arrow_type):
            return "TINYINT"
        if pa.types.is_int16(arrow_type):
            return "SMALLINT"
        if pa.types.is_int32(arrow_type):
            return "INT"
        if pa.types.is_int64(arrow_type):
            return "BIGINT"
        if pa.types.is_float32(arrow_type):
            return "FLOAT"
        if pa.types.is_float64(arrow_type):
            return "DOUBLE"
        if is_arrow_type_string_like(arrow_type):
            return "STRING"
        if is_arrow_type_binary_like(arrow_type):
            return "BINARY"
        if pa.types.is_timestamp(arrow_type):
            tz = getattr(arrow_type, "tz", None)
            return "TIMESTAMP" if tz else "TIMESTAMP_NTZ"
        if pa.types.is_date(arrow_type):
            return "DATE"
        if pa.types.is_decimal(arrow_type):
            return f"DECIMAL({arrow_type.precision}, {arrow_type.scale})"
        if pa.types.is_null(arrow_type):
            # Pragmatic fallback: Delta doesn't support a dedicated NULL column type.
            return "STRING"
        raise ValueError(f"Cannot make ddl type for {arrow_type}")
